/*
 * gdevge.c --- device driver for [CEV]GA (for DJGPP 2.01)
 *
 * Copyright (C) 1997, 1998  Michael Bukin
 */

/*
 * Ghostscript device driver for CGA, EGA and VGA.
 * Mono and 4 bpp color depths.
 * Supports orientation and resolution.
 * Text console over graphics.
 * Compile with DJGPP 2.01.
 */

#include <conio.h>
#include <dpmi.h>
#include <go32.h>
#include <signal.h>
#include <stdlib.h>

#include "string_.h"
#include "gx.h"
#include "gserrors.h"
#include "gsparam.h"
#include "gxdevice.h"
#include "gdevpccm.h"

#include "gp_djcon.h"

/* Device procedures.  */
private dev_proc_open_device (ge_open);
private dev_proc_output_page (ge_output_page);
private dev_proc_close_device (ge_close);
private dev_proc_map_rgb_color (ge_map_rgb_color);
private dev_proc_map_color_rgb (ge_map_color_rgb);
private dev_proc_fill_rectangle (ge_fill_rectangle);
private dev_proc_copy_mono (ge_copy_mono);
private dev_proc_copy_color (ge_copy_color);
private dev_proc_get_params (ge_get_params);
private dev_proc_put_params (ge_put_params);
private dev_proc_get_page_device (ge_get_page_device);

/* Device descriptor structure.  */
typedef struct gx_ge_device_s
{
  gx_device_common;
  int ge_width;		/* width of screen in pixels.  */
  int ge_height;	/* height of screen in pixels.  */
  int ge_depth;		/* depth of screen (bpp).  */
  int ge_is_page_device;	/* is it a page device?  */
  long ge_fg_rgb;	/* console foreground color components.  */
  long ge_bg_rgb;	/* console background color components.  */
  int ge_fg_color;	/* console foreground color.  */
  int ge_bg_color;	/* console background color.  */
} gx_ge_device;

#define ge_dev ((gx_ge_device*) dev)

/* Procedures overriden by this device.  */
private gx_device_procs ge_procs =
{
  ge_open,
  NULL,			/* get_initial_matrix  */
  NULL,			/* sync_output  */
  ge_output_page,
  ge_close,
  ge_map_rgb_color,
  ge_map_color_rgb,
  ge_fill_rectangle,
  NULL,			/* tile_rectangle  */
  ge_copy_mono,
  ge_copy_color,
  NULL,			/* draw_line  */
  NULL,			/* ge_get_bits  */
  ge_get_params,
  ge_put_params,
  NULL,			/* map_cmyk_color  */
  NULL,			/* get_xfont_procs  */
  NULL,			/* get_xfont_device  */
  NULL,			/* map_rgb_alpha_color  */
  ge_get_page_device,
  NULL,			/* get_alpha_bits  */
  NULL			/* copy_alpha  */
};

/* Page height in inches.  */
#define PHI 11.69
/* Support for Orientation.  */
#define bit_options(p) ((p == param_Orientation) ? 0 : -1)

gx_ge_device far_data gs_ge_device =
{
  std_device_full_options_body
  (gx_ge_device,
   &ge_procs,
   "ge",
   640, 480,
   480 / PHI, 480 / PHI,
   3, 4, 2, 2, 3, 3,
   0, 0, 0, 0, 0, 0,
   bit_options),
  { 0 },
  640, 480, 4,
  1,
  255255255,
  0,
  15,
  0
};

/* Forward declarations for console functions.  */
private int geCON_init (int sw, int sh, int pw, int ph,
			int cw, int ch, int fg, int bg);
private void geCON_done (void);
private void geCON_print_char (int ch);
private void geCON_clear_char (void);
private void geCON_blink_char (void);
private void geCON_scroll (void);
private void geCON_update (void);
private void geCON_toggle (void);
private int geCON_is_invalid (void);

/* Bitmap structure.  */
typedef struct GE_BITMAP
{
  int w, h, seg;
  unsigned char* dat;
  unsigned char* line[0];
} GE_BITMAP;

/* Scaled 8x8 font.  */
typedef struct GE_FNT
{
  GE_BITMAP* dat[256];
} GE_FNT;

/* Table of graphics functions.  */
typedef struct GE_VTABLE
{
  GE_BITMAP* (*create_mbitmap) (int w, int h);
  GE_BITMAP* (*create_sbitmap) (int w, int h);
  void (*destroy_bitmap) (GE_BITMAP* bmp);
  void (*putpixel) (GE_BITMAP* bmp, int x, int y, int c);
  int  (*getpixel) (GE_BITMAP* bmp, int x, int y);
  void (*rectfill) (GE_BITMAP* bmp, int x1, int y1, int x2, int y2, int c);
  void (*blit) (GE_BITMAP* src, GE_BITMAP* dst, int sx, int sy,
		int dx, int dy, int w, int h);
} GE_VTABLE;

/* Table of functions currently in use.  */
private GE_VTABLE* ge_vtable = 0;

/* Console for EGA driver.  */
struct geCON_t
{
  djCON_common;
  int fg;
  int bg;
  GE_FNT* font;
  GE_BITMAP* save;
  GE_BITMAP* text;
  GE_BITMAP* page;
};

/* Console instance.  */
private struct geCON_t geCON =
{
  djCON_common_data,
  geCON_print_char,
  geCON_clear_char,
  geCON_blink_char,
  geCON_scroll,
  geCON_update,
  geCON_toggle,
  geCON_is_invalid,
  0, 0, 0, 0, 0, 0
};

/* Modes for ge_set_gfx_mode.  */
#define GE_GFX_TEXT 0
#define GE_GFX_AUTODETECT 1

/* Forward declarations for EGA access functions.  */
private int ge_init (void);
private int ge_set_gfx_mode (int card, int w, int h, int d);

private __inline__ GE_BITMAP*
ge_create_mbitmap (int w, int h)
{
  return ge_vtable->create_mbitmap (w, h);
}
private __inline__ GE_BITMAP*
ge_create_sbitmap (int w, int h)
{
  return ge_vtable->create_sbitmap (w, h);
}
private __inline__ void
ge_destroy_bitmap (GE_BITMAP* bmp)
{
  ge_vtable->destroy_bitmap (bmp);
}
private __inline__ void
ge_putpixel (GE_BITMAP* bmp, int x, int y, int c)
{
  ge_vtable->putpixel (bmp, x, y, c);
}
private __inline__ int
ge_getpixel (GE_BITMAP* bmp, int x, int y)
{
  return ge_vtable->getpixel (bmp, x, y);
}
private __inline__ void
ge_rectfill (GE_BITMAP* bmp, int x1, int y1, int x2, int y2, int c)
{
  ge_vtable->rectfill (bmp, x1, y1, x2, y2, c);
}
private __inline__ void
ge_blit (GE_BITMAP* src, GE_BITMAP* dst, int sx, int sy,
	 int dx, int dy, int w, int h)
{
  ge_vtable->blit (src, dst, sx, sy, dx, dy, w, h);
}
private __inline__ void
ge_clear_to_color (GE_BITMAP* bmp, int c)
{
  ge_vtable->rectfill (bmp, 0, 0, bmp->w - 1, bmp->h - 1, c);
}

/* Convert of color values.  */
#define ge_8bits_to_cv(v) ((v) << (gx_color_value_bits - 8))
#define ge_cv_to_8bits(v) ((v) >> (gx_color_value_bits - 8))

/*
 * Find closest color.
 */
int
ge_closest_color (gx_device* dev, int r, int g, int b)
{
  int i, bi = (int) gx_no_color_index;
  int d, bd = 256 * 256 * 3;
  r = ge_cv_to_8bits (r);
  g = ge_cv_to_8bits (g);
  b = ge_cv_to_8bits (b);

  /* For all colors.  */
  for (i = 0; i < (1 << ge_dev->ge_depth); i++)
    {
      gx_color_value rgb[3];

      ge_map_color_rgb (dev, i, rgb);
      rgb[0] = ge_cv_to_8bits (rgb[0]);
      rgb[1] = ge_cv_to_8bits (rgb[1]);
      rgb[2] = ge_cv_to_8bits (rgb[2]);
      d = ((r - rgb[0]) * (r - rgb[0])
	   + (g - rgb[1]) * (g - rgb[1])
	   + (b - rgb[2]) * (b - rgb[2]));
      if (d < bd)
	{
	  bi = i;
	  bd = d;
	}
    }

  return bi;
}

/*
 * Convert rgb->color for console colors.
 */
void
ge_init_colors (gx_device* dev)
{
  int i, r, g, b;

  /* Allocate foreground color for console.  */
  i = ge_dev->ge_fg_rgb;
  r = ge_8bits_to_cv ((i / 1000000) & 255);
  i %= 1000000;
  g = ge_8bits_to_cv ((i / 1000) & 255);
  i %= 1000;
  b = ge_8bits_to_cv (i & 255);
  ge_dev->ge_fg_color = ge_map_rgb_color (dev, r, g, b);
  if (ge_dev->ge_fg_color == (int) gx_no_color_index)
    ge_dev->ge_fg_color = ge_closest_color (dev, r, g, b);

  /* Allocate background color for console.  */
  i = ge_dev->ge_bg_rgb;
  r = ge_8bits_to_cv ((i / 1000000) & 255);
  i %= 1000000;
  g = ge_8bits_to_cv ((i / 1000) & 255);
  i %= 1000;
  b = ge_8bits_to_cv (i & 255);
  ge_dev->ge_bg_color = ge_map_rgb_color (dev, r, g, b);
  if (ge_dev->ge_bg_color == (int) gx_no_color_index)
    ge_dev->ge_bg_color = ge_closest_color (dev, r, g, b);
}


/*
 * ====== Ghostscript driver functions. ======
 */

/*
 * Initialize device (to graphics mode).
 */
private int
ge_open (gx_device* dev)
{
  int cw, ch;

  /* Initialize graphics lib.  */
  if (ge_init () < 0)
    return_error (gs_error_rangecheck);

  /* Enter graphics mode.  */
  if (ge_set_gfx_mode (GE_GFX_AUTODETECT,
		       ge_dev->ge_width, ge_dev->ge_height,
		       ge_dev->ge_depth) < 0)
    return_error (gs_error_rangecheck);

  /* Convert rgb->color.  */
  ge_init_colors (dev);

  /* Character size.  */
  cw = ch = ge_dev->ge_width / 40;

  /* Initialize console.  */
  if (geCON_init (ge_dev->ge_width, ge_dev->ge_height,
		  dev->width, dev->height, cw, ch,
		  ge_dev->ge_fg_color, ge_dev->ge_bg_color) < 0)
    {
      ge_set_gfx_mode (GE_GFX_TEXT, 0, 0, 0);
      return_error (gs_error_rangecheck);
    }

  /* Capture console.  */
  djCON_capture ((struct djCON_t*) &geCON);

  return 0;
}

/*
 * Output complete page.
 */
private int
ge_output_page (gx_device* dev, int num_copies, int flush)
{
  djCON_update ();
  return 0;
}

/*
 * Shutdown device (to text mode).
 */
private int
ge_close (gx_device* dev)
{
  djCON_release ();
  geCON_done ();
  ge_set_gfx_mode (GE_GFX_TEXT, 0, 0, 0);
  return 0;
}

/*
 * Map rgb to color index.
 */
private gx_color_index
ge_map_rgb_color (gx_device* dev, gx_color_value r,
		  gx_color_value g, gx_color_value b)
{
  if (dev->color_info.depth == 1)
    return gx_default_map_rgb_color (dev, r, g, b);
  else if (dev->color_info.depth == 4)
    return pc_4bit_map_rgb_color (dev, r, g, b);
  return 0;
}

/*
 * Map color index to rgb.
 */
private int
ge_map_color_rgb (gx_device* dev, gx_color_index color,
		  gx_color_value rgb[3])
{
  if (dev->color_info.depth == 1)
    rgb[0] = rgb[1] = rgb[2] = (int) color ? gx_max_color_value : 0;
  else if (dev->color_info.depth == 4)
    return pc_4bit_map_color_rgb (dev, color, rgb);
  return 0;
}

/*
 * Fill a rectangle.
 */
private int
ge_fill_rectangle (gx_device* dev, int x, int y, int w, int h,
		   gx_color_index color)
{
  fit_fill (dev, x, y, w, h);
  if ((w <= 0) || (h <= 0) || (color == gx_no_color_index))
    return 0;
  ge_rectfill (geCON.page, x, y, x + w - 1, y + h - 1, color);
  return 0;
}

/* Frequently used lookup tables.  */
private int ge_bmask[8] = { 0x80, 0x40, 0x20, 0x10,
			    0x08, 0x04, 0x02, 0x01 };
private int ge_nmask[8] = { 0x7F, 0xBF, 0xDF, 0xEF,
			    0xF7, 0xFB, 0xFD, 0xFE };
private int ge_fmask[8] = { 0xFF, 0x7F, 0x3F, 0x1F,
			    0x0F, 0x07, 0x03, 0x01 };
private int ge_lmask[8] = { 0x80, 0xC0, 0xE0, 0xF0,
			    0xF8, 0xFC, 0xFE, 0xFF };

/*
 * Copy bitmap.
 */
private int
ge_copy_mono (gx_device* dev, const byte* base, int sourcex, int sraster,
	      gx_bitmap_id id, int x, int y, int w, int h,
	      gx_color_index zero, gx_color_index one)
{
  int xi;
  int yi;
  int sw;
  int xs;
  int yh;

  fit_copy (dev, base, sourcex, sraster, id, x, y, w, h);
  if ((w <= 0) || (h <= 0))
    return 0;

  sw = sourcex + w;
  xs = x - sourcex;
  yh = y + h;
  if (zero == gx_no_color_index)
    {
      if (one == gx_no_color_index)
	{
	  /* Nothing to change.  */
	  return 0;
	}
      else
	{
	  /* Write only bits with 1.  */
	  for (yi = y; yi < yh; yi++, base += sraster)
	    for (xi = sourcex; xi < sw; xi++)
	      if (base[xi >> 3] & ge_bmask[xi & 7])
		ge_putpixel (geCON.page, xi + xs, yi, (int) one);
	}
    }
  else
    {
      if (one == gx_no_color_index)
	{
	  /* Write only bits with 0.  */
	  for (yi = y; yi < yh; yi++, base += sraster)
	    for (xi = sourcex; xi < sw; xi++)
	      if (!(base[xi >> 3] & ge_bmask[xi & 7]))
		ge_putpixel (geCON.page, xi + xs, yi, (int) zero);
	}
      else
	{
	  /* Copy bitmap.  */
	  for (yi = y; yi < yh; yi++, base += sraster)
	    for (xi = sourcex; xi < sw; xi++)
	      if (base[xi >> 3] & ge_bmask[xi & 7])
		ge_putpixel (geCON.page, xi + xs, yi, (int) one);
	      else
		ge_putpixel (geCON.page, xi + xs, yi, (int) zero);
	}
    }
  return 0;
}

/*
 * Copy pixmap.
 */
private int
ge_copy_color (gx_device* dev, const byte* base, int sourcex, int sraster,
	       gx_bitmap_id id, int x, int y, int w, int h)
{
  int xi;
  int yi;
  int sw;
  int xs;
  int yh;

  fit_copy (dev, base, sourcex, sraster, id, x, y, w, h);
  if ((w <= 0) || (h <= 0))
    return 0;

  sw = sourcex + w;
  xs = x - sourcex;
  yh = y + h;
  if (dev->color_info.depth == 1)
    {
      for (yi = y; yi < yh; yi++, base += sraster)
	for (xi = sourcex; xi < sw; xi++)
	  if (base[xi >> 3] & ge_bmask[xi & 7])
	    ge_putpixel (geCON.page, xi + xs, yi, 1);
	  else
	    ge_putpixel (geCON.page, xi + xs, yi, 0);
    }
  else if (dev->color_info.depth == 4)
    {
      for (yi = y; yi < yh; yi++, base += sraster)
	for (xi = sourcex; xi < sw; xi++)
	  if (xi & 1)
	    ge_putpixel (geCON.page, xi + xs, yi,
			 (int) base[xi >> 1] & 0x0F);
	  else
	    ge_putpixel (geCON.page, xi + xs, yi,
			 (int) base[xi >> 1] >> 4);
    }
  return 0;
}

/*
 * Get parameters.
 */
private int
ge_get_params (gx_device* dev, gs_param_list* list)
{
  int code;

  code = gx_default_get_params (dev, list);
  if (code < 0)
    return code;

  code = param_write_int (list, "ScreenWidth", &ge_dev->ge_width);
  if (code < 0)
    return code;

  code = param_write_int (list, "ScreenHeight", &ge_dev->ge_height);
  if (code < 0)
    return code;

  code = param_write_int (list, "ScreenDepth", &ge_dev->ge_depth);
  if (code < 0)
    return code;

  code = param_write_long (list, "ConsoleForeground", &ge_dev->ge_fg_rgb);
  if (code < 0)
    return code;

  code = param_write_long (list, "ConsoleBackground", &ge_dev->ge_bg_rgb);
  if (code < 0)
    return code;

  code = param_write_int (list, "IsPageDevice", &ge_dev->ge_is_page_device);
  if (code < 0)
    return code;

  return code;
}

/*
 * Put parameters. Defined parameters:
 * ScreenWidth (int), ScreenHeight (int), ScreenDepth (int),
 * ConsoleForeground (long), ConsoleBackground (long),
 * IsPageDevice (int).
 */
private int
ge_put_params (gx_device* dev, gs_param_list* list)
{
  int code;
  int ecode = 0;
  int width = ge_dev->ge_width;
  int height = ge_dev->ge_height;
  int depth = ge_dev->ge_depth;
  long fg_rgb = ge_dev->ge_fg_rgb;
  long bg_rgb = ge_dev->ge_bg_rgb;
  int is_page_device = ge_dev->ge_is_page_device;
  static gx_device_color_info ge_1bit_color = dci_black_and_white;
  static gx_device_color_info ge_4bit_color = dci_pc_4bit;
  const char* param_name;

  code = param_read_int (list, (param_name = "ScreenWidth"), &width);
  if ((code != 0) && (code != 1))
    param_signal_error (list, param_name, (ecode = code));

  code = param_read_int (list, (param_name = "ScreenHeight"), &height);
  if ((code != 0) && (code != 1))
    param_signal_error (list, param_name, (ecode = code));

  code = param_read_int (list, (param_name = "ScreenDepth"), &depth);
  if ((code == 0) && (depth != 1) && (depth != 4))
    param_signal_error (list, param_name, (ecode = gs_error_rangecheck));
  else if ((code != 0) && (code != 1))
    param_signal_error (list, param_name, (ecode = code));

  code = param_read_long (list, (param_name = "ConsoleForeground"), &fg_rgb);
  if ((code != 0) && (code != 1))
    param_signal_error (list, param_name, (ecode = code));
  
  code = param_read_long (list, (param_name = "ConsoleBackground"), &bg_rgb);
  if ((code != 0) && (code != 1))
    param_signal_error (list, param_name, (ecode = code));

  code = param_read_int (list, (param_name = "IsPageDevice"), &is_page_device);
  if ((code != 0) && (code != 1))
    param_signal_error (list, param_name, (ecode = code));

  code = gx_default_put_params (dev, list);

  if (ecode < 0)
    return ecode;
  if (code < 0)
    return code;

  ge_dev->ge_width = width;
  ge_dev->ge_height = height;
  ge_dev->ge_depth = depth;
  ge_dev->ge_fg_rgb = fg_rgb;
  ge_dev->ge_bg_rgb = bg_rgb;
  ge_dev->ge_is_page_device = is_page_device;

  switch (depth)
    {
    case 1:
      dev->color_info = ge_1bit_color;
      break;
    case 4:
      dev->color_info = ge_4bit_color;
      break;
    }

  return 0;
}

/*
 * Get page device.
 */
private gx_device*
ge_get_page_device (gx_device* dev)
{
  return ((ge_dev->ge_is_page_device) ? dev : 0);
}


/*
 * ====== EGA access functions. ======
 */

/*
 * This is optimized for speed for given task.
 * It should be correct for all Ghostscript use of graphics.
 * Some restrictions:
 *  1. screen bitmaps must have left side aligned on 8 pixels.
 *  2. when blitting one part of bitmap to another part of the same
 *     bitmap, sy must be less than dy (blit).
 *  3. screen bitmaps should have width of screen.
 *  4. bmp->seg must be the same for all screen bitmaps (s->s blit).
 *  5. screen width must be <= 800 pixels (m->s blit).
 *  6. graphics controller registers must be set as following:
 *     0x0000, 0x0001, 0x0003, 0x0205.
 *     this way there is no need to setup them for put/getpixel.
 *  7. all bitmaps must have the same color depth.
 *  ...
 */

/* Forward declarations for low level functions.  */
private GE_BITMAP* ge_real_create_mbitmap1 (int w, int h);
private GE_BITMAP* ge_real_create_mbitmap4 (int w, int h);
private GE_BITMAP* ge_real_create_sbitmap1 (int w, int h);
private GE_BITMAP* ge_real_create_sbitmap1_cga (int w, int h);
private GE_BITMAP* ge_real_create_sbitmap4 (int w, int h);
private void ge_real_destroy_bitmap (GE_BITMAP* bmp);
private void ge_real_putpixel1 (GE_BITMAP* bmp, int x, int y, int c);
private void ge_real_putpixel4 (GE_BITMAP* bmp, int x, int y, int c);
private int  ge_real_getpixel1 (GE_BITMAP* bmp, int x, int y);
private int  ge_real_getpixel4 (GE_BITMAP* bmp, int x, int y);
private void ge_real_rectfill1 (GE_BITMAP* bmp, int x1, int y1,
				int x2, int y2, int c);
private void ge_real_rectfill4 (GE_BITMAP* bmp, int x1, int y1,
				int x2, int y2, int c);
private void ge_real_blit1 (GE_BITMAP* src, GE_BITMAP* dst, int sx, int sy,
			    int dx, int dy, int w, int h);
private void ge_real_blit4 (GE_BITMAP* src, GE_BITMAP* dst, int sx, int sy,
			    int dx, int dy, int w, int h);

private GE_VTABLE ge_vtable1 =
{
  ge_real_create_mbitmap1,
  ge_real_create_sbitmap1,
  ge_real_destroy_bitmap,
  ge_real_putpixel1,
  ge_real_getpixel1,
  ge_real_rectfill1,
  ge_real_blit1
};

private GE_VTABLE ge_vtable1_cga =
{
  ge_real_create_mbitmap1,
  ge_real_create_sbitmap1_cga,
  ge_real_destroy_bitmap,
  ge_real_putpixel1,
  ge_real_getpixel1,
  ge_real_rectfill1,
  ge_real_blit1
};

private GE_VTABLE ge_vtable4 =
{
  ge_real_create_mbitmap4,
  ge_real_create_sbitmap4,
  ge_real_destroy_bitmap,
  ge_real_putpixel4,
  ge_real_getpixel4,
  ge_real_rectfill4,
  ge_real_blit4
};

private struct
{
  int m, w, h, d;
  GE_VTABLE* vtable;
} ge_mode_info[] =
{
  { 0x06, 640, 200, 1, &ge_vtable1_cga },
  { 0x0F, 640, 350, 1, &ge_vtable1 },
  { 0x11, 640, 480, 1, &ge_vtable1 },
  { 0x12, 640, 480, 4, &ge_vtable4 },
  { 0x10, 640, 350, 4, &ge_vtable4 },
  { 0x0E, 640, 200, 4, &ge_vtable4 },
  { 0x0D, 320, 200, 4, &ge_vtable4 },
  { 0, 0, 0, 0, 0 }
};

private int ge_count = 0;

private int ge_gfx_virgin = 1;
private int ge_a_rez = 3;
private int ge_a_lines = -1;

private int ge_call_shutdown_gfx = 0;

private int ge_color_depth = 4;
private GE_BITMAP* ge_screen = 0;

private void
ge_check_gfx_virginity (void)
{
  if (ge_gfx_virgin)
    {
      __dpmi_regs r;
      struct text_info textinfo;

      /* Get previous video mode.  */
      memset (&r, 0, sizeof (r));
      r.x.ax = 0x0F00;
      __dpmi_int (0x10, &r);
      ge_a_rez = r.x.ax & 0xFF;

      if (ge_a_rez == 3)
	{
	  /* Get number of screen lines.  */
	  gettextinfo (&textinfo);
	  ge_a_lines = textinfo.screenheight;
	}
      else
	ge_a_lines = -1;

      ge_call_shutdown_gfx = 1;

      ge_gfx_virgin = 0;
    }
}

private int
ge_set_gfx_mode (int card, int w, int h, int d)
{
  ge_check_gfx_virginity ();

  if (ge_screen != 0)
    {
      ge_destroy_bitmap (ge_screen);
      ge_screen = 0;
      ge_vtable = 0;
    }

  if (card == GE_GFX_TEXT)
    {
      __dpmi_regs r;

      /* Set text mode.  */
      memset (&r, 0, sizeof (r));
      r.x.ax = ge_a_rez;
      __dpmi_int (0x10, &r);

      if (ge_a_lines > 0)
	_set_screen_lines (ge_a_lines);

      return 0;
    }
  else
    {
      int i;
      __dpmi_regs r;

      /* Find video mode.  */
      for (i = 0; ge_mode_info[i].w != 0; i++)
	if ((ge_mode_info[i].w == w)
	    && (ge_mode_info[i].h == h)
	    && (ge_mode_info[i].d == d))
	  break;

      if (ge_mode_info[i].w == 0)
	return -1;

      /* Set video mode.  */
      memset (&r, 0, sizeof (r));
      r.x.ax = ge_mode_info[i].m;
      __dpmi_int (0x10, &r);

      /* Test video mode.  */
      memset (&r, 0, sizeof (r));
      r.x.ax = 0x0F00;
      __dpmi_int (0x10, &r);
      if ((r.x.ax & 0xFF) != ge_mode_info[i].m)
	return -1;

      /* Save mode parameters.  */
      ge_vtable = ge_mode_info[i].vtable;
      ge_color_depth = d;

      /* Create screen bitmap.  */
      ge_screen = ge_create_sbitmap (w, h);
      if (ge_screen == 0)
	{
	  ge_vtable = 0;
	  return -1;
	}

      return 0;
    }
}

private void
ge_shutdown_gfx (void)
{
  __dpmi_regs r;

  /* Get current video mode.  */
  memset (&r, 0, sizeof (r));
  r.x.ax = 0x0F00;
  __dpmi_int (0x10, &r);

  /* Restore previous video mode.  */
  if (((r.x.ax & 0xFF) != ge_a_rez) || (ge_screen != 0))
    ge_set_gfx_mode (GE_GFX_TEXT, 0, 0, 0);
}

private void
ge_exit (void)
{
  if (ge_call_shutdown_gfx)
    ge_shutdown_gfx ();
}

private void
ge_signal_handler (int num)
{
  ge_exit ();
  signal (num, SIG_DFL);
  raise (num);
}

private int
ge_init (void)
{
  if (ge_count == 0)
    atexit (ge_exit);

  signal (SIGABRT, ge_signal_handler);
  signal (SIGFPE, ge_signal_handler);
  signal (SIGILL, ge_signal_handler);
  signal (SIGSEGV, ge_signal_handler);
  signal (SIGTERM, ge_signal_handler);
  signal (SIGINT, ge_signal_handler);
  signal (SIGKILL, ge_signal_handler);
  signal (SIGQUIT, ge_signal_handler);

  ge_count++;
  return 0;
}

private GE_BITMAP*
ge_real_create_mbitmap1 (int w, int h)
{
  int i;
  unsigned char* d;
  GE_BITMAP* b;

  if ((w < 0) || (h < 0))
    return 0;

#define width_fixup(w) (((w) + 7) & ~7)
  b = (GE_BITMAP*) malloc (sizeof (GE_BITMAP) + sizeof (unsigned char*) * h);
  d = (unsigned char*) malloc (width_fixup (w) * h / 8);
  if ((b == 0) || (d == 0))
    {
      if (b != 0)
	free (b);
      if (d != 0)
	free (d);
      return 0;
    }
  for (i = 0; i < h; i++)
    b->line[i] = d + i * width_fixup (w) / 8;
  b->w = w;
  b->h = h;
  b->dat = d;
  b->seg = -1;
#undef width_fixup

  return b;
}

private GE_BITMAP*
ge_real_create_mbitmap4 (int w, int h)
{
  int i;
  unsigned char* d;
  GE_BITMAP* b;

  if ((w < 0) || (h < 0))
    return 0;

  b = (GE_BITMAP*) malloc (sizeof (GE_BITMAP) + sizeof (unsigned char*) * h);
  d = (unsigned char*) malloc (w * h);
  if ((b == 0) || (d == 0))
    {
      if (b != 0)
	free (b);
      if (d != 0)
	free (d);
      return 0;
    }
  for (i = 0; i < h; i++)
    b->line[i] = d + i * w;
  b->w = w;
  b->h = h;
  b->dat = d;
  b->seg = -1;

  return b;
}

private GE_BITMAP*
ge_real_create_sbitmap1 (int w, int h)
{
  int i;
  GE_BITMAP* b;

  b = (GE_BITMAP*) malloc (sizeof (GE_BITMAP) + sizeof (unsigned char*) * h);
  if (b == 0)
    return 0;
  for (i = 0; i < h; i++)
    b->line[i] = (unsigned char*) (0xA0000 + i * w / 8);
  b->w = w;
  b->h = h;
  b->dat = 0;
  b->seg = _dos_ds;

  return b;
}

private GE_BITMAP*
ge_real_create_sbitmap1_cga (int w, int h)
{
  int i;
  GE_BITMAP* b;

  b = (GE_BITMAP*) malloc (sizeof (GE_BITMAP) + sizeof (unsigned char*) * h);
  if (b == 0)
    return 0;
  for (i = 0; i < h; i++)
    if (i & 1)
      b->line[i] = (unsigned char*) (0xBA000 + (i / 2) * w / 8);
    else
      b->line[i] = (unsigned char*) (0xB8000 + (i / 2) * w / 8);
  b->w = w;
  b->h = h;
  b->dat = 0;
  b->seg = _dos_ds;

  return b;
}

private GE_BITMAP*
ge_real_create_sbitmap4 (int w, int h)
{
  int i;
  GE_BITMAP* b;

  b = (GE_BITMAP*) malloc (sizeof (GE_BITMAP) + sizeof (unsigned char*) * h);
  if (b == 0)
    return 0;
  for (i = 0; i < h; i++)
    b->line[i] = (unsigned char*) (0xA0000 + (i * w) / 8);
  b->w = w;
  b->h = h;
  b->dat = 0;
  b->seg = _dos_ds;

  /* Prepare for put/getpixel.  */
  dj_outportw (0x3CE, 0x0205);

  return b;
}

private void
ge_real_destroy_bitmap (GE_BITMAP* bmp)
{
  if (bmp != 0)
    {
      if (bmp->dat != 0)
	{
	  free (bmp->dat);
	  bmp->dat = 0;
	}
      free (bmp);
    }
}

private void
ge_real_putpixel1 (GE_BITMAP* bmp, int x, int y, int c)
{
  if ((x < 0) || (x >= bmp->w) || (y < 0) || (y >= bmp->h))
    return;
  if (bmp->dat != 0)
    {
      /*
       * Memory.
       */
      if (c)
	bmp->line[y][x >> 3] |= ge_bmask[x & 7];
      else
	bmp->line[y][x >> 3] &= ge_nmask[x & 7];
    }
  else
    {
      /*
       * Screen.
       */
      int d = (int) bmp->line[y] + (x >> 3);

      _dj_farsetsel (bmp->seg);
      if (c)
	_dj_farnspokeb (d, _dj_farnspeekb (d) | ge_bmask[x & 7]);
      else
	_dj_farnspokeb (d, _dj_farnspeekb (d) & ge_nmask[x & 7]);
    }
}

private void
ge_real_putpixel4 (GE_BITMAP* bmp, int x, int y, int c)
{
  if ((x < 0) || (x >= bmp->w) || (y < 0) || (y >= bmp->h))
    return;
  if (bmp->dat != 0)
    {
      /*
       * Memory.
       */
      bmp->line[y][x] = c;
    }
  else
    {
      /*
       * Screen.
       */
      int d = (int) bmp->line[y] + (x >> 3);

      dj_outportw (0x3CE, (ge_bmask[x & 7] << 8) | 0x0008);
      _dj_farsetsel (bmp->seg);
      _dj_farnspeekb (d);
      _dj_farnspokeb (d, c);
    }
}

private int
ge_real_getpixel1 (GE_BITMAP* bmp, int x, int y)
{
  if ((x < 0) || (x >= bmp->w) || (y < 0) || (y >= bmp->h))
    return 0;
  if (bmp->dat != 0)
    {
      /*
       * Memory.
       */
      return (bmp->line[y][x >> 3] & ge_bmask[x & 7]) ? 1 : 0;
    }
  else
    {
      /*
       * Screen.
       */
      int d = (int) bmp->line[y] + (x >> 3);

      return (_dj_farpeekb (bmp->seg, d) & ge_bmask[x & 7]) ? 1 : 0;
    }
}

private int
ge_real_getpixel4 (GE_BITMAP* bmp, int x, int y)
{
  if ((x < 0) || (x >= bmp->w) || (y < 0) || (y >= bmp->h))
    return 0;
  if (bmp->dat != 0)
    {
      /*
       * Memory.
       */
      return bmp->line[y][x];
    }
  else
    {
      /*
       * Screen.
       */
      int d = (int) bmp->line[y] + (x >> 3);
      int b = ge_bmask[x & 7];
      int v = 0;

      _dj_farsetsel (bmp->seg);
#define add_bit(i) if ((_dj_farnspeekb (d) & b) != 0) v |= (i)
      dj_outportw (0x3CE, 0x0004);
      add_bit (0x01);
      dj_outportw (0x3CE, 0x0104);
      add_bit (0x02);
      dj_outportw (0x3CE, 0x0204);
      add_bit (0x04);
      dj_outportw (0x3CE, 0x0304);
      add_bit (0x08);
#undef add_bit
      return v;
    }
}

private void
ge_real_rectfill1 (GE_BITMAP* bmp, int x1, int y1, int x2, int y2, int c)
{
  int y, dfm, dlm;
  int x1i, x1s, x2i, x2s;

  if (x1 > x2)
    {
      int t = x1;
      x1 = x2;
      x2 = t;
    }
  if (x1 < 0)
    x1 = 0;
  if (x2 >= bmp->w)
    x2 = bmp->w - 1;
  if (x1 > x2)
    return;

  if (y1 > y2)
    {
      int t = y1;
      y1 = y2;
      y2 = t;
    }
  if (y1 < 0)
    y1 = 0;
  if (y2 >= bmp->h)
    y2 = bmp->h - 1;
  if (y1 > y2)
    return;

  x1i = x1 >> 3;
  x1s = x1 & 7;
  x2i = x2 >> 3;
  x2s = x2 & 7;

  if (bmp->dat != 0)
    {
      /*
       * Memory.
       */
      if (c != 0)
	{
	  dfm = ge_fmask[x1s];
	  dlm = ge_lmask[x2s];
	  if (x1i == x2i)
	    {
	      dfm &= dlm;
	      dlm = 0x00;
	    }
	  for (y = y1; y <= y2; y++)
	    {
	      unsigned char* df = bmp->line[y] + x1i;
	      unsigned char* dl = bmp->line[y] + x2i;
	      unsigned char* d;

	      *df = *df | dfm;
	      for (d = (df + 1); d < dl; d++)
		*d = 0xFF;
	      *dl = *dl | dlm;
	    }
	}
      else
	{
	  dfm = ~ge_fmask[x1s];
	  dlm = ~ge_lmask[x2s];
	  if (x1i == x2i)
	    {
	      dfm |= dlm;
	      dlm = 0xFF;
	    }
	  for (y = y1; y <= y2; y++)
	    {
	      unsigned char* df = bmp->line[y] + x1i;
	      unsigned char* dl = bmp->line[y] + x2i;
	      unsigned char* d;

	      *df = *df & dfm;
	      for (d = (df + 1); d < dl; d++)
		*d = 0x00;
	      *dl = *dl & dlm;
	    }
	}
    }
  else
    {
      /*
       * Screen.
       */
      _dj_farsetsel (bmp->seg);
      if (c != 0)
	{
	  dfm = ge_fmask[x1s];
	  dlm = ge_lmask[x2s];
	  if (x1i == x2i)
	    {
	      dfm &= dlm;
	      dlm = 0x00;
	    }
	  for (y = y1; y <= y2; y++)
	    {
	      int df = (int) bmp->line[y] + x1i;
	      int dl = (int) bmp->line[y] + x2i;
	      int d;

	      _dj_farnspokeb (df, _dj_farnspeekb (df) | dfm);
	      for (d = (df + 1); d < dl; d++)
		_dj_farnspokeb (d, 0xFF);
	      _dj_farnspokeb (dl, _dj_farnspeekb (dl) | dlm);
	    }
	}
      else
	{
	  dfm = ~ge_fmask[x1s];
	  dlm = ~ge_lmask[x2s];
	  if (x1i == x2i)
	    {
	      dfm |= dlm;
	      dlm = 0xFF;
	    }
	  for (y = y1; y <= y2; y++)
	    {
	      int df = (int) bmp->line[y] + x1i;
	      int dl = (int) bmp->line[y] + x2i;
	      int d;

	      _dj_farnspokeb (df, _dj_farnspeekb (df) & dfm);
	      for (d = (df + 1); d < dl; d++)
		_dj_farnspokeb (d, 0x00);
	      _dj_farnspokeb (dl, _dj_farnspeekb (dl) & dlm);
	    }
	}
    }
}

private void
ge_real_rectfill4 (GE_BITMAP* bmp, int x1, int y1, int x2, int y2, int c)
{
  if (x1 > x2)
    {
      int t = x1;
      x1 = x2;
      x2 = t;
    }
  if (x1 < 0)
    x1 = 0;
  if (x2 >= bmp->w)
    x2 = bmp->w - 1;
  if (x1 > x2)
    return;

  if (y1 > y2)
    {
      int t = y1;
      y1 = y2;
      y2 = t;
    }
  if (y1 < 0)
    y1 = 0;
  if (y2 >= bmp->h)
    y2 = bmp->h - 1;
  if (y1 > y2)
    return;

  if (bmp->dat != 0)
    {
      /*
       * Memory.
       */
      int x, y;

      for (y = y1; y <= y2; y++)
	{
	  unsigned char* d = bmp->line[y];
	  for (x = x1; x <= x2; x++)
	    d[x] = c;
	}
    }
  else
    {
      /*
       * Screen.
       */
      int b, y;
      int x1i, x1s;
      int x2i, x2s;

      x1i = x1 >> 3;
      x1s = x1 & 7;
      x2i = x2 >> 3;
      x2s = x2 & 7;

      _dj_farsetsel (bmp->seg);

      /* First nibble.  */
      if ((x1s != 0) || (x1i == x2i))
	{
	  b = ge_fmask[x1s];
	  if (x1i == x2i)
	    b &= ge_lmask[x2s];
	  dj_outportw (0x3CE, (b << 8) | 0x0008);
	  for (y = y1; y <= y2; y++)
	    {
	      int d = (int) bmp->line[y] + x1i;

	      _dj_farnspeekb (d);
	      _dj_farnspokeb (d, c);
	    }
	  if (x1i == x2i)
	    return;
	  x1i++;
	}

      /* Last nibble.  */
      if (x2s != 7)
	{
	  b = ge_lmask[x2s];
	  dj_outportw (0x3CE, (b << 8) | 0x0008);
	  for (y = y1; y <= y2; y++)
	    {
	      int d = (int) bmp->line[y] + x2i;

	      _dj_farnspeekb (d);
	      _dj_farnspokeb (d, c);
	    }
	  x2i--;
	}

      if (x1i > x2i)
	return;

      /* Full bytes.  */
      dj_outportw (0x3CE, 0xFF08);
      _dj_farnspokeb ((int) bmp->line[y1] + x1i, c);
      _dj_farnspeekb ((int) bmp->line[y1] + x1i);
      dj_outportw (0x3CE, 0x0105);
      for (y = y1; y <= y2; y++)
	{
	  int df = (int) bmp->line[y] + x1i;
	  int dl = (int) bmp->line[y] + x2i;
	  int d;

	  for (d = df; d <= dl; d++)
	    _dj_farnspokeb (d, c);
	}
      dj_outportw (0x3CE, 0x0205);
    }
}

private void
ge_real_blit1 (GE_BITMAP* src, GE_BITMAP* dst,
	       int sx, int sy, int dx, int dy, int w, int h)
{
  int x, y;

  x = (sx < dx) ? sx : dx;
  if (x < 0)
    {
      w += x;
      sx -= x;
      dx -= x;
    }
  y = (sy < dy) ? sy : dy;
  if (y < 0)
    {
      h += y;
      sy -= y;
      dy -= y;
    }
  if ((w <= 0) || (h <= 0))
    return;

  x = ((src->w - sx) < (dst->w - dx)) ? (src->w - sx) : (dst->w - dx);
  if (w > x)
    w = x;
  y = ((src->h - sy) < (dst->w - dy)) ? (src->h - sy) : (dst->h - dy);
  if (h > y)
    h = y;
  if ((w <= 0) || (h <= 0))
    return;

  if ((src->dat != 0) && (dst->dat != 0))
    {
      /*
       * Memory->memory blit.
       */
      int s1i, s1s, s2i, s2s;
      int d1i, d1s, d2i, d2s;
      int sfm, slm, dfm, dlm;
      int sds, sdy, val, yh;

      yh = dy + h;
      sdy = sy - dy;
      s1i = sx >> 3;
      s1s = sx & 7;
      s2i = (sx + w - 1) >> 3;
      s2s = (sx + w - 1) & 7;
      d1i = dx >> 3;
      d1s = dx & 7;
      d2i = (dx + w - 1) >> 3;
      d2s = (dx + w - 1) & 7;
      sds = (s1s > d1s) ? (8 - s1s + d1s) : (d1s - s1s);

      sfm = ge_fmask[d1s];
      dfm = ~sfm;
      slm = ge_lmask[d2s];
      dlm = ~slm;
      if (d1i == d2i)
	{
	  sfm &= slm;
	  dfm = ~sfm;
	  slm = 0x00;
	  dlm = 0xFF;
	}

      if (s1s == d1s)
	{
	  /* No shift.  */
	  for (y = dy; y < yh; y++)
	    {
	      unsigned char* df = dst->line[y] + d1i;
	      unsigned char* dl = dst->line[y] + d2i;
	      unsigned char* d;
	      int sdi = (int) src->line[y + sdy] + s1i - (int) df;

	      *df = (*df & dfm) | (*(df + sdi) & sfm);
	      for (d = df + 1; d < dl; d++)
		*d = *(d + sdi);
	      *dl = (*dl & dlm) | (*(dl + sdi) & slm);
	    }
	}
      else if (s1s < d1s)
	{
	  /* Shift source right.  */
	  for (y = dy; y < yh; y++)
	    {
	      unsigned char* df = dst->line[y] + d1i;
	      unsigned char* dl = dst->line[y] + d2i;
	      unsigned char* d;
	      int sdi = (int) src->line[y + sdy] + s1i - (int) df;

	      val = *(df + sdi);
	      *df = (*df & dfm) | ((val >> sds) & sfm);
	      val <<= 8;
	      for (d = df + 1; d < dl; d++)
		{
		  val |= *(d + sdi);
		  *d = val >> sds;
		  val <<= 8;
		}
	      if (s2s < d2s)
		val |= *(dl + sdi);
	      *dl = (*dl & dlm) | ((val >> sds) & slm);
	    }
	}
      else /* s1s > d1s  */
	{
	  /* Shift source left.  */
	  for (y = dy; y < yh; y++)
	    {
	      unsigned char* df = dst->line[y] + d1i;
	      unsigned char* dl = dst->line[y] + d2i;
	      unsigned char* d;
	      int sdi = (int) src->line[y + sdy] + s1i - (int) df;

	      val = *(dl + sdi) << 8;
	      if (s2s < d2s)
		val |= *(dl + 1 + sdi);
	      *dl = (*dl & dlm) | ((val >> sds) & slm);
	      val >>= 8;
	      for (d = dl - 1; d > df; d--)
		{
		  val |= *(d + sdi) << 8;
		  *d = val >> sds;
		  val >>= 8;
		}
	      val |= *(df + sdi) << 8;
	      *df = (*df & dfm) | ((val >> sds) & sfm);
	    }
	}
    }
  else if ((src->dat != 0) && (dst->dat == 0))
    {
      /*
       * Memory->screen blit.
       */
      int s1i, s1s, s2i, s2s;
      int d1i, d1s, d2i, d2s;
      int sfm, slm, dfm, dlm;
      int sds, sdy, val, yh;

      yh = dy + h;
      sdy = sy - dy;
      s1i = sx >> 3;
      s1s = sx & 7;
      s2i = (sx + w - 1) >> 3;
      s2s = (sx + w - 1) & 7;
      d1i = dx >> 3;
      d1s = dx & 7;
      d2i = (dx + w - 1) >> 3;
      d2s = (dx + w - 1) & 7;
      sds = (s1s > d1s) ? (8 - s1s + d1s) : (d1s - s1s);

      sfm = ge_fmask[d1s];
      dfm = ~sfm;
      slm = ge_lmask[d2s];
      dlm = ~slm;
      if (d1i == d2i)
	{
	  sfm &= slm;
	  dfm = ~sfm;
	  slm = 0x00;
	  dlm = 0xFF;
	}

      _dj_farsetsel (dst->seg);
      if (s1s == d1s)
	{
	  /* No shift.  */
	  for (y = dy; y < yh; y++)
	    {
	      int df = (int) dst->line[y] + d1i;
	      int dl = (int) dst->line[y] + d2i;
	      int d;
	      unsigned char* sdi = src->line[y + sdy] + s1i - df;

	      _dj_farnspokeb (df, ((_dj_farnspeekb (df) & dfm)
				   | (*(df + sdi) & sfm)));
	      for (d = df + 1; d < dl; d++)
		_dj_farnspokeb (d, *(d + sdi));
	      _dj_farnspokeb (dl, ((_dj_farnspeekb (dl) & dlm)
				   | (*(dl + sdi) & slm)));
	    }
	}
      else if (s1s < d1s)
	{
	  /* Shift source right.  */
	  for (y = dy; y < yh; y++)
	    {
	      int df = (int) dst->line[y] + d1i;
	      int dl = (int) dst->line[y] + d2i;
	      int d;
	      unsigned char* sdi = src->line[y + sdy] + s1i - df;

	      val = *(df + sdi);
	      _dj_farnspokeb (df, ((_dj_farnspeekb (df) & dfm)
				   | ((val >> sds) & sfm)));
	      val <<= 8;
	      for (d = df + 1; d < dl; d++)
		{
		  val |= *(d + sdi);
		  _dj_farnspokeb (d, val >> sds);
		  val <<= 8;
		}
	      if (s2s < d2s)
		val |= *(dl + sdi);
	      _dj_farnspokeb (dl, ((_dj_farnspeekb (dl) & dlm)
				   | ((val >> sds) & slm)));
	    }
	}
      else /* s1s > d1s  */
	{
	  /* Shift source left.  */
	  for (y = dy; y < yh; y++)
	    {
	      int df = (int) dst->line[y] + d1i;
	      int dl = (int) dst->line[y] + d2i;
	      int d;
	      unsigned char* sdi = src->line[y + sdy] + s1i - df;

	      val = *(dl + sdi) << 8;
	      if (s2s < d2s)
		val |= *(dl + 1 + sdi);
	      _dj_farnspokeb (dl, ((_dj_farnspeekb (dl) & dlm)
				   | ((val >> sds) & slm)));
	      val >>= 8;
	      for (d = dl - 1; d > df; d--)
		{
		  val |= *(d + sdi) << 8;
		  _dj_farnspokeb (d, val >> sds);
		  val >>= 8;
		}
	      val |= *(df + sdi) << 8;
	      _dj_farnspokeb (df, ((_dj_farnspeekb (df) & dfm)
				   | ((val >> sds) & sfm)));
	    }
	}
    }
  else
    {
      /*
       * Leftovers.
       */
      int xw, yh;
      int sdx, sdy;

      xw = dx + w;
      yh = dy + h;
      sdx = sx - dx;
      sdy = sy - dy;

      for (y = dy; y < yh; y++)
	for (x = dx; x < xw; x++)
	  ge_putpixel (dst, x, y, ge_getpixel (src, x + sdx, y + sdy));
    }
}

private void
ge_real_blit4 (GE_BITMAP* src, GE_BITMAP* dst,
	       int sx, int sy, int dx, int dy, int w, int h)
{
  int x, y;

  x = (sx < dx) ? sx : dx;
  if (x < 0)
    {
      w += x;
      sx -= x;
      dx -= x;
    }
  y = (sy < dy) ? sy : dy;
  if (y < 0)
    {
      h += y;
      sy -= y;
      dy -= y;
    }
  if ((w <= 0) || (h <= 0))
    return;

  x = ((src->w - sx) < (dst->w - dx)) ? (src->w - sx) : (dst->w - dx);
  if (w > x)
    w = x;
  y = ((src->h - sy) < (dst->w - dy)) ? (src->h - sy) : (dst->h - dy);
  if (h > y)
    h = y;
  if ((w <= 0) || (h <= 0))
    return;

  if ((src->dat != 0) && (dst->dat != 0))
    {
      /*
       * Memory->memory blit.
       */
      int xw, yh;
      int sdy, sdx;

      xw = dx + w;
      yh = dy + h;
      sdx = sx - dx;
      sdy = sy - dy;

      for (y = dy; y < yh; y++)
	{
	  unsigned char* df = dst->line[y] + dx;
	  unsigned char* dl = dst->line[y] + xw;
	  unsigned char* d;
	  int s = (int) src->line[y + sdy] + sx - (int) df;

	  for (d = df; d < dl; d++)
	    *d = *(s + d);
	}
    }
  else if ((src->dat != 0) && (dst->dat == 0))
    {
      /*
       * Memory->screen blit.
       */
      static unsigned int mask[8][16] =
      {
	{
	  0x00000000, 0x00000080, 0x00008000, 0x00008080,
	  0x00800000, 0x00800080, 0x00808000, 0x00808080,
	  0x80000000, 0x80000080, 0x80008000, 0x80008080,
	  0x80800000, 0x80800080, 0x80808000, 0x80808080
	},
	{
	  0x00000000, 0x00000040, 0x00004000, 0x00004040,
	  0x00400000, 0x00400040, 0x00404000, 0x00404040,
	  0x40000000, 0x40000040, 0x40004000, 0x40004040,
	  0x40400000, 0x40400040, 0x40404000, 0x40404040
	},
	{
	  0x00000000, 0x00000020, 0x00002000, 0x00002020,
	  0x00200000, 0x00200020, 0x00202000, 0x00202020,
	  0x20000000, 0x20000020, 0x20002000, 0x20002020,
	  0x20200000, 0x20200020, 0x20202000, 0x20202020
	},
	{
	  0x00000000, 0x00000010, 0x00001000, 0x00001010,
	  0x00100000, 0x00100010, 0x00101000, 0x00101010,
	  0x10000000, 0x10000010, 0x10001000, 0x10001010,
	  0x10100000, 0x10100010, 0x10101000, 0x10101010
	},
	{
	  0x00000000, 0x00000008, 0x00000800, 0x00000808,
	  0x00080000, 0x00080008, 0x00080800, 0x00080808,
	  0x08000000, 0x08000008, 0x08000800, 0x08000808,
	  0x08080000, 0x08080008, 0x08080800, 0x08080808
	},
	{
	  0x00000000, 0x00000004, 0x00000400, 0x00000404,
	  0x00040000, 0x00040004, 0x00040400, 0x00040404,
	  0x04000000, 0x04000004, 0x04000400, 0x04000404,
	  0x04040000, 0x04040004, 0x04040400, 0x04040404
	},
	{
	  0x00000000, 0x00000002, 0x00000200, 0x00000202,
	  0x00020000, 0x00020002, 0x00020200, 0x00020202,
	  0x02000000, 0x02000002, 0x02000200, 0x02000202,
	  0x02020000, 0x02020002, 0x02020200, 0x02020202
	},
	{
	  0x00000000, 0x00000001, 0x00000100, 0x00000101,
	  0x00010000, 0x00010001, 0x00010100, 0x00010101,
	  0x01000000, 0x01000001, 0x01000100, 0x01000101,
	  0x01010000, 0x01010001, 0x01010100, 0x01010101
	}
      };
      static unsigned int v[100];
      int d1i, d2i, sdy;
      int sfm, dfm, slm, dlm;
      int yh, xw;

      _dj_farsetsel (dst->seg);
      dj_outportw (0x3CE, 0x0005);
      dj_outportw (0x3CE, 0xFF08);

      xw = dx + w;
      yh = dy + h;
      sdy = sy - dy;
      d1i = dx >> 3;
      d2i = (xw - 1) >> 3;
      sfm = ge_fmask[dx & 7];
      dfm = ~sfm;
      slm = ge_lmask[(xw - 1) & 7];
      dlm = ~slm;
      if (d1i == d2i)
	{
	  sfm &= slm;
	  dfm = ~sfm;
	  slm = 0x00;
	  dlm = 0xFF;
	}

      for (y = dy; y < yh; y++)
	{
	  int df = (int) dst->line[y] + d1i;
	  int dl = (int) dst->line[y] + d2i;
	  unsigned int*  vdi = v - df;
	  unsigned char* sdi = src->line[y + sdy] + sx - dx;

	  for (x = df; x <= dl; x++)
	    *(vdi + x) = 0;

	  for (x = dx; x < xw; x++)
	    *(v + (x >> 3) - d1i) |= mask[x & 7][*(sdi + x) & 0x0F];

	  dj_outportw (0x3C4, 0x0102);
	  dj_outportw (0x3CE, 0x0004);
	  _dj_farnspokeb (df, ((*((unsigned char*) (vdi + df)) & sfm)
			       | (_dj_farnspeekb (df) & dfm)));
	  for (x = df + 1; x < dl; x++)
	    _dj_farnspokeb (x, *((unsigned char*) (vdi + x)));
	  _dj_farnspokeb (dl, ((*((unsigned char*) (vdi + dl)) & slm)
			       | (_dj_farnspeekb (dl) & dlm)));

	  dj_outportw (0x3C4, 0x0202);
	  dj_outportw (0x3CE, 0x0104);
	  _dj_farnspokeb (df, ((*((unsigned char*) (vdi + df) + 1) & sfm)
			       | (_dj_farnspeekb (df) & dfm)));
	  for (x = df + 1; x < dl; x++)
	    _dj_farnspokeb (x, *((unsigned char*) (vdi + x) + 1));
	  _dj_farnspokeb (dl, ((*((unsigned char*) (vdi + dl) + 1) & slm)
			       | (_dj_farnspeekb (dl) & dlm)));

	  dj_outportw (0x3C4, 0x0402);
	  dj_outportw (0x3CE, 0x0204);
	  _dj_farnspokeb (df, ((*((unsigned char*) (vdi + df) + 2) & sfm)
			       | (_dj_farnspeekb (df) & dfm)));
	  for (x = df + 1; x < dl; x++)
	    _dj_farnspokeb (x, *((unsigned char*) (vdi + x) + 2));
	  _dj_farnspokeb (dl, ((*((unsigned char*) (vdi + dl) + 2) & slm)
			       | (_dj_farnspeekb (dl) & dlm)));

	  dj_outportw (0x3C4, 0x0802);
	  dj_outportw (0x3CE, 0x0304);
	  _dj_farnspokeb (df, ((*((unsigned char*) (vdi + df) + 3) & sfm)
			       | (_dj_farnspeekb (df) & dfm)));
	  for (x = df + 1; x < dl; x++)
	    _dj_farnspokeb (x, *((unsigned char*) (vdi + x) + 3));
	  _dj_farnspokeb (dl, ((*((unsigned char*) (vdi + dl) + 3) & slm)
			       | (_dj_farnspeekb (dl) & dlm)));
	}
      dj_outportw (0x3CE, 0x0205);
      dj_outportw (0x3C4, 0x0F02);
    }
  else
    {
      /*
       * Leftovers.
       */
      int xw, yh;
      int sdy, sdx;

      xw = dx + w;
      yh = dy + h;
      sdx = sx - dx;
      sdy = sy - dy;

      for (y = dy; y < yh; y++)
	for (x = dx; x < xw; x++)
	  ge_putpixel (dst, x, y, ge_getpixel (src, x + sdx, y + sdy));
    }
}


/*
 * ====== Console functions. ======
 */

private GE_FNT*
geCON_create_font (int cw, int ch, int fg, int bg)
{
  int i;
  GE_FNT* f;
  GE_BITMAP* b;
  int x, x1, x2;
  int y, y1, y2;

  if ((cw < 8) || (ch < 8))
    return 0;
  
  f = (GE_FNT*) malloc (sizeof (GE_FNT));
  if (f == 0)
    return 0;

  for (i = 0; i < 256; i++)
    {
      f->dat[i] = b = ge_create_mbitmap (cw, ch);
      if (b == 0)
	continue;

      y2 = 0;
      for (y = 0; y < 8; y++)
	{
	  y1 = y2;
	  y2 = (ch * (y + 1)) / 8;
	  x2 = 0;
	  for (x = 0; x < 8; x++)
	    {
	      x1 = x2;
	      x2 = (cw * (x + 1)) / 8;
	      if (djCON_font_data[i][y] & (0x80 >> x))
		ge_rectfill (b, x1, y1, x2 - 1, y2 - 1, fg);
	      else
		ge_rectfill (b, x1, y1, x2 - 1, y2 - 1, bg);
	    }
	}
    }

  return f;
}

private void
geCON_destroy_font (GE_FNT* f)
{
  int i;

  if (f != 0)
    {
      for (i = 0; i < 256; i++)
	if (f->dat[i] != 0)
	  {
	    ge_destroy_bitmap (f->dat[i]);
	    f->dat[i] = 0;
	  }
      free (f);
    }
}

private int
geCON_init (int sw, int sh, int pw, int ph, int cw, int ch, int fg, int bg)
{
  geCON.sw = sw;
  geCON.sh = sh;
  geCON.pw = pw;
  geCON.ph = ph;
  geCON.vph = sh;
  geCON.tw = sw;
  geCON.th = sh / 2;
  geCON.vth = 0;
  geCON.cw = cw;
  geCON.ch = ch;
  geCON.cx = 0;
  geCON.cy = 0;
  geCON.ox = 0;
  geCON.oy = 0;
  geCON.lr = ((geCON.sw / 64) != 0) ? (geCON.sw / 64) : 5;
  geCON.ud = ((geCON.sh / 48) != 0) ? (geCON.sh / 48) : 5;
  geCON.fg = fg;
  geCON.bg = bg;
  geCON.font = geCON_create_font (cw, ch, fg, bg);
  geCON.save = ge_create_mbitmap (geCON.tw, geCON.th);
  geCON.text = ge_create_sbitmap (geCON.tw, geCON.th);
  geCON.page = ge_create_mbitmap (geCON.pw, geCON.ph);
  if ((geCON.font == 0) || (geCON.save == 0)
      || (geCON.text == 0) || (geCON.page == 0))
    {
      geCON_done ();
      return -1;
    }
  ge_clear_to_color (geCON.save, geCON.bg);
  return 0;
}

private void
geCON_done (void)
{
  if (geCON.font != 0)
    {
      geCON_destroy_font (geCON.font);
      geCON.font = 0;
    }
  if (geCON.save != 0)
    {
      ge_destroy_bitmap (geCON.save);
      geCON.save = 0;
    }
  if (geCON.text != 0)
    {
      ge_destroy_bitmap (geCON.text);
      geCON.text = 0;
    }
  if (geCON.page != 0)
    {
      ge_destroy_bitmap (geCON.page);
      geCON.page = 0;
    }
}

private void
geCON_print_char (int ch)
{
  if (geCON.font->dat[ch] == 0)
    return;

  ge_blit (geCON.font->dat[ch], geCON.save, 0, 0,
	   geCON.cx, geCON.cy, geCON.cw, geCON.ch);

  if (geCON.vth != 0)
    ge_blit (geCON.font->dat[ch], geCON.text, 0, 0,
	     geCON.cx, geCON.cy, geCON.cw, geCON.ch);
}

private void
geCON_clear_char (void)
{
  ge_rectfill (geCON.save, geCON.cx, geCON.cy,
	       geCON.cx + geCON.cw - 1, geCON.cy + geCON.ch - 1, geCON.bg);

  if (geCON.vth != 0)
    ge_rectfill (geCON.text, geCON.cx, geCON.cy,
		 geCON.cx + geCON.cw - 1, geCON.cy + geCON.ch - 1, geCON.bg);
}

private void
geCON_blink_char (void)
{
  int x, y;

  for (y = geCON.cy; y < (geCON.cy + geCON.ch); y++)
    for (x = geCON.cx; x < (geCON.cx + geCON.cw); x++)
      if (ge_getpixel (geCON.save, x, y) == geCON.bg)
	ge_putpixel (geCON.save, x, y, geCON.fg);
      else
	ge_putpixel (geCON.save, x, y, geCON.bg);

  if (geCON.vth != 0)
    ge_blit (geCON.save, geCON.text, geCON.cx, geCON.cy,
	     geCON.cx, geCON.cy, geCON.cw, geCON.ch);
}

private void
geCON_scroll (void)
{
  ge_blit (geCON.save, geCON.save, 0, geCON.ch, 0, 0,
	   geCON.tw, geCON.th - geCON.ch);
  ge_rectfill (geCON.save, 0, geCON.th - geCON.ch,
	       geCON.tw - 1, geCON.th - 1, geCON.bg);

  if (geCON.vth != 0)
    ge_blit (geCON.save, geCON.text, 0, 0, 0, 0, geCON.tw, geCON.th);
}

private void
geCON_update (void)
{
  ge_blit (geCON.page, ge_screen, geCON.ox, geCON.oy + geCON.vth,
	   0, geCON.vth, geCON.sw, geCON.vph);
  if ((geCON.pw - geCON.ox) < geCON.sw)
    ge_rectfill (ge_screen, geCON.pw - geCON.ox, geCON.vth,
		 geCON.sw - 1, geCON.sh - 1, geCON.fg);
  if ((geCON.ph - geCON.oy) <= geCON.vth)
    ge_rectfill (ge_screen, 0, geCON.vth,
		 geCON.pw - 1, geCON.sh - 1, geCON.fg);
  else if ((geCON.ph - geCON.oy) < geCON.sh)
    ge_rectfill (ge_screen, 0, geCON.ph - geCON.oy,
		 geCON.pw - 1, geCON.sh - 1, geCON.fg);
}

private void
geCON_toggle (void)
{
  if (geCON.vth != 0)
    ge_blit (geCON.save, geCON.text, 0, 0, 0, 0, geCON.tw, geCON.th);
  else
    geCON_update ();
}

private int
geCON_is_invalid (void)
{
  return ((ge_screen == 0)
	  || (geCON.font == 0)
	  || (geCON.save == 0)
	  || (geCON.text == 0)
	  || (geCON.page == 0));
}

/*
 * gdevge.c ends here
 */
