/*
 * gp_djcon.c --- text console over graphics
 *
 * Copyright (C) 1998  Michael Bukin
 */

/*
 * Text console over graphics.
 * Compile with DJGPP 2.01.
 *
 * Ghostscript naming convention is broken purposedly (djCON_).
 */

#include <conio.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/fsext.h>

#include "gp_djcon.h"

/* Currently used instance of console.  */
static struct djCON_t* djCON = 0;

/*
 * Move cursor position to the start of the next line.
 * Scrolling text when at bottom of console.
 */
static void
djCON_forward_line (void)
{
  djCON->cx = 0;
  djCON->cy += djCON->ch;
  if ((djCON->cy + djCON->ch) <= djCON->th)
    return;
  djCON->cy -= djCON->ch;
  djCON->scroll ();
}

/*
 * Insert character and move to the next position.
 */
static void
djCON_insert_char (int ch)
{
  djCON->print_char (ch);
  djCON->cx += djCON->cw;
  if ((djCON->cx + djCON->cw) > djCON->sw)
    djCON_forward_line ();
}

/*
 * Move cursor position to the start of line.
 */
static void
djCON_beginning_of_line (void)
{
  djCON->cx = 0;
}

/*
 * Delete character before cursor.
 */
static void
djCON_backward_delete_char (void)
{
  djCON->cx -= djCON->cw;
  if ((djCON->cx < 0) && (djCON->cy > 0))
    {
      djCON->cy -= djCON->ch;
      djCON->cx = djCON->tw - (djCON->tw % djCON->cw) - djCON->cw;
      if (djCON->cx < 0)
	djCON->cx = 0;
      if (djCON->cy < 0)
	djCON->cy = 0;
    }
  djCON->clear_char ();
}

/*
 * Return nonzero when at the top-left corner of console.
 */
static int
djCON_at_the_beginning (void)
{
  return ((djCON->cx == 0) && (djCON->cy == 0));
}

/*
 * Update console.
 */
void
djCON_update (void)
{
  if ((djCON == 0) || (djCON->is_invalid ()))
    return;
  djCON->update ();
}

/*
 * Toggle console on-off screen.
 */
static void
djCON_toggle (void)
{
  if (djCON->vth == 0)
    {
      /* Show console.  */
      djCON->vth = djCON->th;
    }
  else
    {
      /* Hide console.  */
      djCON->vth = 0;
    }
  djCON->vph = djCON->sh - djCON->vth;
  djCON->toggle ();
}

/*
 * Move page left on specified amount of pixels.
 */
static void
djCON_page_move_left (const int amount)
{
  djCON->ox += amount;
  if ((djCON->ox < 0) || (djCON->pw < djCON->sw))
    djCON->ox = 0;
  else if ((djCON->pw - djCON->ox) < djCON->sw)
    djCON->ox = djCON->pw - djCON->sw;
}

/*
 * Move page up on specified amount of pixels.
 */
static void
djCON_page_move_up (const int amount)
{
  djCON->oy += amount;
  if ((djCON->oy < 0) || (djCON->ph < djCON->sh))
    djCON->oy = 0;
  else if ((djCON->ph - djCON->oy) < djCON->sh)
    djCON->oy = djCON->ph - djCON->sh;
}

/*
 * Print character on console.
 */
static void
djCON_putc (int ch)
{
  switch (ch)
    {
    case '\n':
      djCON_forward_line ();
      break;
    case '\r':
      djCON_beginning_of_line ();
      break;
    case '\b':
      djCON_backward_delete_char ();
      break;
    default:
      djCON_insert_char (ch);
      break;
    }
}

/*
 * Print len-string on console.
 */
static int
djCON_puts (const char* str, const int len)
{
  int i;

  if ((djCON == 0) || (djCON->is_invalid ()))
    return len;

  for (i = 0; i < len; i++)
    djCON_putc (str[i]);

  return len;
}

/*
 * Get character from console.
 * Also handle specially some keys.
 */
static int
djCON_getc (void)
{
  while (1)
    {
      int ch;

      /* Get key.  */
      ch = getch ();
      if (ch != 0)
	{
	  /* Not extended character.  */
	  switch (ch)
	    {
	    case 27:
	      /* <esc>.  */
	      djCON_toggle ();
	      break;
	    case '\r':
	      /* <ret>.  */
	      return '\n';
	    case '\t':
	      /* <tab>.  */
	      return ' ';
	    case '\b':
	      /* <backspace>.  */
	      return '\b';
	    default:
	      if (ch >= 32)
		return ch;
	      break;
	    }
	  continue;
	}

      /* Extended character.  */
      ch = getch ();
      switch (ch)
	{
	case 83:
	  /* <delete>.  */
	  return '\b';
	case 75:
	  /* <left>.  */
	  djCON_page_move_left (-djCON->lr);
	  break;
	case 77:
	  /* <right>.  */
	  djCON_page_move_left (djCON->lr);
	  break;
	case 72:
	  /* <up>.  */
	  djCON_page_move_up (-djCON->ud);
	  break;
	case 80:
	  /* <down>.  */
	  djCON_page_move_up (djCON->ud);
	  break;
	case 155:
	  /* M-<left>.  */
	  djCON_page_move_left (-1);
	  break;
	case 157:
	  /* M-<right>.  */
	  djCON_page_move_left (1);
	  break;
	case 152:
	  /* M-<up>.  */
	  djCON_page_move_up (-1);
	  break;
	case 160:
	  /* M-<down>.  */
	  djCON_page_move_up (1);
	  break;
	case 115:
	  /* C-<left>.  */
	  djCON_page_move_left (-djCON->sw / 2);
	  break;
	case 116:
	  /* C-<right>.  */
	  djCON_page_move_left (djCON->sw / 2);
	  break;
	case 141:
	  /* C-<up>.  */
	  djCON_page_move_up (-djCON->vph / 2);
	  break;
	case 145:
	  /* C-<down>.  */
	  djCON_page_move_up (djCON->vph / 2);
	  break;
	case 71:
	  /* <home>.  */
	  djCON_page_move_left (-djCON->pw);
	  break;
	case 79:
	  /* <end>.  */
	  djCON_page_move_left (djCON->pw);
	  break;
	case 119:
	  /* C-<home>.  */
	  djCON_page_move_up (-djCON->ph);
	  break;
	case 117:
	  /* C-<end>.  */
	  djCON_page_move_up (djCON->ph);
	  break;
	case 73:
	  /* <prior>.  */
	  djCON_page_move_up (-djCON->vph / 2);
	  break;
	case 81:
	  /* <next>.  */
	  djCON_page_move_up (djCON->vph / 2);
	  break;
	default:
	  continue;
	}
      djCON->update ();
    }
}

#define djCON_BUFFER_SIZE 4096

/*
 * Get string from console.
 */
static int
djCON_gets (char* str, const int len)
{
  static int buflen = 0;
  static char buf[djCON_BUFFER_SIZE];

  if ((djCON == 0) || (djCON->is_invalid ()))
    return 0;

  if (buflen > 0)
    {
      /* Have data from previous call.  */
      if (len < buflen)
	{
	  /* Data fit in buffer.  */
	  memcpy (str, buf, len);
	  memcpy (buf, buf + len, buflen - len);
	  buflen -= len;
	  return len;
	}
      else
	{
	  /* Data does not fit in buffer.  */
	  int oldlen;
	  memcpy (str, buf, buflen);
	  oldlen = buflen;
	  buflen = 0;
	  return oldlen;
	}
    }

  while (1)
    {
      int ch;

      /* Highligh cursor position.  */
      djCON->blink_char ();
      /* Wait for key.  */
      ch = djCON_getc ();
      /* Unhilight cursor position.  */
      djCON->blink_char ();

      /* Parse entered character.  */
      if (ch == '\b')
	{
	  /* Backspace or delete.  */
	  if (buflen <= 0)
	    {
	      /* Already at the beginning of buffer.  */
	      continue;
	    }
	  else if (djCON_at_the_beginning ())
	    {
	      /* On the top of text window.  */
	      djCON_puts (buf, buflen);
	    }
	  /* Delete character from buffer and screen.  */
	  buflen--;
	  djCON_putc ('\b');
	}
      else
	{
	  /* Insert character into buffer.  */
	  if (buflen >= sizeof (buf))
	    {
	      /* Replace last entered character.  */
	      djCON_putc ('\b');
	      buflen = sizeof (buf) - 1;
	    }
	  /* Add character to buffer.  */
	  buf[buflen++] = ch;
	  djCON_putc (ch);
	  if (ch == '\n')
	    {
	      /* String accepted.  */
	      if (len < buflen)
		{
		  /* Data fit in buffer.  */
		  memcpy (str, buf, len);
		  memcpy (buf, buf + len, buflen - len);
		  buflen -= len;
		  return len;
		}
	      else
		{
		  /* Data does not fit in buffer.  */
		  int oldlen;
		  memcpy (str, buf, buflen);
		  oldlen = buflen;
		  buflen = 0;
		  return oldlen;
		}
	    }
	}
    }
}

/*
 * Handler for input from console.
 */
static int
djCON_input_handler (__FSEXT_Fnumber func, int* retval, va_list rest)
{
  int fd;
  int len;
  char* buf;

  fd = va_arg (rest, int);
  if ((djCON == 0) || (djCON->is_invalid ())
      || (func != __FSEXT_read) || (!isatty (fd)))
    return 0;

  buf = va_arg (rest, char*);
  len = va_arg (rest, size_t);
  *retval = djCON_gets (buf, len);

  return 1;
}

/*
 * Handler for output to console.
 */
static int
djCON_output_handler (__FSEXT_Fnumber func, int* retval, va_list rest)
{
  int fd;
  int len;
  char* buf;

  fd = va_arg (rest, int);
  if ((djCON == 0) || (djCON->is_invalid ())
      || (func != __FSEXT_write))
    return 0;

  buf = va_arg (rest, char*);
  len = va_arg (rest, size_t);
  *retval = djCON_puts (buf, len);

  return isatty (fd);
}

/*
 * Capture standard input and output to console.
 */
void
djCON_capture (struct djCON_t* console)
{
  djCON = console;

  if (djCON != 0)
    {
      /* Replace file operation handlers.  */
      __FSEXT_set_function (fileno (stdin), djCON_input_handler);
      __FSEXT_set_function (fileno (stdout), djCON_output_handler);
      __FSEXT_set_function (fileno (stderr), djCON_output_handler);
    }
}

/*
 * Release captured input and output.
 */
void
djCON_release (void)
{
  djCON = 0;
}

/*
 * Console font data.
 */
char djCON_font_data[256][8] =
{
  { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
  { 0x7E, 0x81, 0xA5, 0x81, 0xBD, 0x99, 0x81, 0x7E },
  { 0x7E, 0xFF, 0xDB, 0xFF, 0xC3, 0xE7, 0xFF, 0x7E },
  { 0x6C, 0xFE, 0xFE, 0xFE, 0x7C, 0x38, 0x10, 0x00 },
  { 0x10, 0x38, 0x7C, 0xFE, 0x7C, 0x38, 0x10, 0x00 },
  { 0x38, 0x7C, 0x38, 0xFE, 0xFE, 0xD6, 0x10, 0x38 },
  { 0x10, 0x38, 0x7C, 0xFE, 0xFE, 0x7C, 0x10, 0x38 },
  { 0x00, 0x00, 0x18, 0x3C, 0x3C, 0x18, 0x00, 0x00 },
  { 0xFF, 0xFF, 0xE7, 0xC3, 0xC3, 0xE7, 0xFF, 0xFF },
  { 0x00, 0x3C, 0x66, 0x42, 0x42, 0x66, 0x3C, 0x00 },
  { 0xFF, 0xC3, 0x99, 0xBD, 0xBD, 0x99, 0xC3, 0xFF },
  { 0x0F, 0x07, 0x0F, 0x7D, 0xCC, 0xCC, 0xCC, 0x78 },
  { 0x3C, 0x66, 0x66, 0x66, 0x3C, 0x18, 0x7E, 0x18 },
  { 0x3F, 0x33, 0x3F, 0x30, 0x30, 0x70, 0xF0, 0xE0 },
  { 0x7F, 0x63, 0x7F, 0x63, 0x63, 0x67, 0xE6, 0xC0 },
  { 0x18, 0xDB, 0x3C, 0xE7, 0xE7, 0x3C, 0xDB, 0x18 },
  { 0x80, 0xE0, 0xF8, 0xFE, 0xF8, 0xE0, 0x80, 0x00 },
  { 0x02, 0x0E, 0x3E, 0xFE, 0x3E, 0x0E, 0x02, 0x00 },
  { 0x18, 0x3C, 0x7E, 0x18, 0x18, 0x7E, 0x3C, 0x18 },
  { 0x66, 0x66, 0x66, 0x66, 0x66, 0x00, 0x66, 0x00 },
  { 0x7F, 0xDB, 0xDB, 0x7B, 0x1B, 0x1B, 0x1B, 0x00 },
  { 0x3E, 0x61, 0x3C, 0x66, 0x66, 0x3C, 0x86, 0x7C },
  { 0x00, 0x00, 0x00, 0x00, 0x7E, 0x7E, 0x7E, 0x00 },
  { 0x18, 0x3C, 0x7E, 0x18, 0x7E, 0x3C, 0x18, 0xFF },
  { 0x18, 0x3C, 0x7E, 0x18, 0x18, 0x18, 0x18, 0x00 },
  { 0x18, 0x18, 0x18, 0x18, 0x7E, 0x3C, 0x18, 0x00 },
  { 0x00, 0x18, 0x0C, 0xFE, 0x0C, 0x18, 0x00, 0x00 },
  { 0x00, 0x30, 0x60, 0xFE, 0x60, 0x30, 0x00, 0x00 },
  { 0x00, 0x00, 0xC0, 0xC0, 0xC0, 0xFE, 0x00, 0x00 },
  { 0x00, 0x24, 0x66, 0xFF, 0x66, 0x24, 0x00, 0x00 },
  { 0x00, 0x18, 0x3C, 0x7E, 0xFF, 0xFF, 0x00, 0x00 },
  { 0x00, 0xFF, 0xFF, 0x7E, 0x3C, 0x18, 0x00, 0x00 },
  { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
  { 0x18, 0x3C, 0x3C, 0x18, 0x18, 0x00, 0x18, 0x00 },
  { 0x66, 0x66, 0x24, 0x00, 0x00, 0x00, 0x00, 0x00 },
  { 0x6C, 0x6C, 0xFE, 0x6C, 0xFE, 0x6C, 0x6C, 0x00 },
  { 0x18, 0x3E, 0x60, 0x3C, 0x06, 0x7C, 0x18, 0x00 },
  { 0x00, 0xC6, 0xCC, 0x18, 0x30, 0x66, 0xC6, 0x00 },
  { 0x38, 0x6C, 0x38, 0x76, 0xDC, 0xCC, 0x76, 0x00 },
  { 0x18, 0x18, 0x30, 0x00, 0x00, 0x00, 0x00, 0x00 },
  { 0x0C, 0x18, 0x30, 0x30, 0x30, 0x18, 0x0C, 0x00 },
  { 0x30, 0x18, 0x0C, 0x0C, 0x0C, 0x18, 0x30, 0x00 },
  { 0x00, 0x66, 0x3C, 0xFF, 0x3C, 0x66, 0x00, 0x00 },
  { 0x00, 0x18, 0x18, 0x7E, 0x18, 0x18, 0x00, 0x00 },
  { 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x30 },
  { 0x00, 0x00, 0x00, 0x7E, 0x00, 0x00, 0x00, 0x00 },
  { 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x18, 0x00 },
  { 0x06, 0x0C, 0x18, 0x30, 0x60, 0xC0, 0x80, 0x00 },
  { 0x7C, 0xC6, 0xCE, 0xD6, 0xE6, 0xC6, 0x7C, 0x00 },
  { 0x18, 0x38, 0x18, 0x18, 0x18, 0x18, 0x7E, 0x00 },
  { 0x7C, 0xC6, 0x06, 0x1C, 0x30, 0x66, 0xFE, 0x00 },
  { 0x7C, 0xC6, 0x06, 0x3C, 0x06, 0xC6, 0x7C, 0x00 },
  { 0x1C, 0x3C, 0x6C, 0xCC, 0xFE, 0x0C, 0x1E, 0x00 },
  { 0xFE, 0xC0, 0xC0, 0xFC, 0x06, 0xC6, 0x7C, 0x00 },
  { 0x38, 0x60, 0xC0, 0xFC, 0xC6, 0xC6, 0x7C, 0x00 },
  { 0xFE, 0xC6, 0x0C, 0x18, 0x30, 0x30, 0x30, 0x00 },
  { 0x7C, 0xC6, 0xC6, 0x7C, 0xC6, 0xC6, 0x7C, 0x00 },
  { 0x7C, 0xC6, 0xC6, 0x7E, 0x06, 0x0C, 0x78, 0x00 },
  { 0x00, 0x18, 0x18, 0x00, 0x00, 0x18, 0x18, 0x00 },
  { 0x00, 0x18, 0x18, 0x00, 0x00, 0x18, 0x18, 0x30 },
  { 0x06, 0x0C, 0x18, 0x30, 0x18, 0x0C, 0x06, 0x00 },
  { 0x00, 0x00, 0x7E, 0x00, 0x00, 0x7E, 0x00, 0x00 },
  { 0x60, 0x30, 0x18, 0x0C, 0x18, 0x30, 0x60, 0x00 },
  { 0x7C, 0xC6, 0x0C, 0x18, 0x18, 0x00, 0x18, 0x00 },
  { 0x7C, 0xC6, 0xDE, 0xDE, 0xDE, 0xC0, 0x78, 0x00 },
  { 0x38, 0x6C, 0xC6, 0xFE, 0xC6, 0xC6, 0xC6, 0x00 },
  { 0xFC, 0x66, 0x66, 0x7C, 0x66, 0x66, 0xFC, 0x00 },
  { 0x3C, 0x66, 0xC0, 0xC0, 0xC0, 0x66, 0x3C, 0x00 },
  { 0xF8, 0x6C, 0x66, 0x66, 0x66, 0x6C, 0xF8, 0x00 },
  { 0xFE, 0x62, 0x68, 0x78, 0x68, 0x62, 0xFE, 0x00 },
  { 0xFE, 0x62, 0x68, 0x78, 0x68, 0x60, 0xF0, 0x00 },
  { 0x3C, 0x66, 0xC0, 0xC0, 0xCE, 0x66, 0x3A, 0x00 },
  { 0xC6, 0xC6, 0xC6, 0xFE, 0xC6, 0xC6, 0xC6, 0x00 },
  { 0x3C, 0x18, 0x18, 0x18, 0x18, 0x18, 0x3C, 0x00 },
  { 0x1E, 0x0C, 0x0C, 0x0C, 0xCC, 0xCC, 0x78, 0x00 },
  { 0xE6, 0x66, 0x6C, 0x78, 0x6C, 0x66, 0xE6, 0x00 },
  { 0xF0, 0x60, 0x60, 0x60, 0x62, 0x66, 0xFE, 0x00 },
  { 0xC6, 0xEE, 0xFE, 0xFE, 0xD6, 0xC6, 0xC6, 0x00 },
  { 0xC6, 0xE6, 0xF6, 0xDE, 0xCE, 0xC6, 0xC6, 0x00 },
  { 0x7C, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0x7C, 0x00 },
  { 0xFC, 0x66, 0x66, 0x7C, 0x60, 0x60, 0xF0, 0x00 },
  { 0x7C, 0xC6, 0xC6, 0xC6, 0xC6, 0xCE, 0x7C, 0x0E },
  { 0xFC, 0x66, 0x66, 0x7C, 0x6C, 0x66, 0xE6, 0x00 },
  { 0x7C, 0xC6, 0x60, 0x38, 0x0C, 0xC6, 0x7C, 0x00 },
  { 0x7E, 0x7E, 0x5A, 0x18, 0x18, 0x18, 0x3C, 0x00 },
  { 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0x7C, 0x00 },
  { 0xC6, 0xC6, 0xC6, 0xC6, 0xC6, 0x6C, 0x38, 0x00 },
  { 0xC6, 0xC6, 0xC6, 0xD6, 0xD6, 0xFE, 0x6C, 0x00 },
  { 0xC6, 0xC6, 0x6C, 0x38, 0x6C, 0xC6, 0xC6, 0x00 },
  { 0x66, 0x66, 0x66, 0x3C, 0x18, 0x18, 0x3C, 0x00 },
  { 0xFE, 0xC6, 0x8C, 0x18, 0x32, 0x66, 0xFE, 0x00 },
  { 0x3C, 0x30, 0x30, 0x30, 0x30, 0x30, 0x3C, 0x00 },
  { 0xC0, 0x60, 0x30, 0x18, 0x0C, 0x06, 0x02, 0x00 },
  { 0x3C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x3C, 0x00 },
  { 0x10, 0x38, 0x6C, 0xC6, 0x00, 0x00, 0x00, 0x00 },
  { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF },
  { 0x30, 0x18, 0x0C, 0x00, 0x00, 0x00, 0x00, 0x00 },
  { 0x00, 0x00, 0x78, 0x0C, 0x7C, 0xCC, 0x76, 0x00 },
  { 0xE0, 0x60, 0x7C, 0x66, 0x66, 0x66, 0xDC, 0x00 },
  { 0x00, 0x00, 0x7C, 0xC6, 0xC0, 0xC6, 0x7C, 0x00 },
  { 0x1C, 0x0C, 0x7C, 0xCC, 0xCC, 0xCC, 0x76, 0x00 },
  { 0x00, 0x00, 0x7C, 0xC6, 0xFE, 0xC0, 0x7C, 0x00 },
  { 0x3C, 0x66, 0x60, 0xF8, 0x60, 0x60, 0xF0, 0x00 },
  { 0x00, 0x00, 0x76, 0xCC, 0xCC, 0x7C, 0x0C, 0xF8 },
  { 0xE0, 0x60, 0x6C, 0x76, 0x66, 0x66, 0xE6, 0x00 },
  { 0x18, 0x00, 0x38, 0x18, 0x18, 0x18, 0x3C, 0x00 },
  { 0x06, 0x00, 0x06, 0x06, 0x06, 0x66, 0x66, 0x3C },
  { 0xE0, 0x60, 0x66, 0x6C, 0x78, 0x6C, 0xE6, 0x00 },
  { 0x38, 0x18, 0x18, 0x18, 0x18, 0x18, 0x3C, 0x00 },
  { 0x00, 0x00, 0xEC, 0xFE, 0xD6, 0xD6, 0xD6, 0x00 },
  { 0x00, 0x00, 0xDC, 0x66, 0x66, 0x66, 0x66, 0x00 },
  { 0x00, 0x00, 0x7C, 0xC6, 0xC6, 0xC6, 0x7C, 0x00 },
  { 0x00, 0x00, 0xDC, 0x66, 0x66, 0x7C, 0x60, 0xF0 },
  { 0x00, 0x00, 0x76, 0xCC, 0xCC, 0x7C, 0x0C, 0x1E },
  { 0x00, 0x00, 0xDC, 0x76, 0x60, 0x60, 0xF0, 0x00 },
  { 0x00, 0x00, 0x7E, 0xC0, 0x7C, 0x06, 0xFC, 0x00 },
  { 0x30, 0x30, 0xFC, 0x30, 0x30, 0x36, 0x1C, 0x00 },
  { 0x00, 0x00, 0xCC, 0xCC, 0xCC, 0xCC, 0x76, 0x00 },
  { 0x00, 0x00, 0xC6, 0xC6, 0xC6, 0x6C, 0x38, 0x00 },
  { 0x00, 0x00, 0xC6, 0xD6, 0xD6, 0xFE, 0x6C, 0x00 },
  { 0x00, 0x00, 0xC6, 0x6C, 0x38, 0x6C, 0xC6, 0x00 },
  { 0x00, 0x00, 0xC6, 0xC6, 0xC6, 0x7E, 0x06, 0xFC },
  { 0x00, 0x00, 0x7E, 0x4C, 0x18, 0x32, 0x7E, 0x00 },
  { 0x0E, 0x18, 0x18, 0x70, 0x18, 0x18, 0x0E, 0x00 },
  { 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x00 },
  { 0x70, 0x18, 0x18, 0x0E, 0x18, 0x18, 0x70, 0x00 },
  { 0x76, 0xDC, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
  { 0x00, 0x10, 0x38, 0x6C, 0xC6, 0xC6, 0xFE, 0x00 },
  { 0x78, 0xCC, 0xC0, 0xCC, 0x78, 0x18, 0x0C, 0x78 },
  { 0x00, 0xCC, 0x00, 0xCC, 0xCC, 0xCC, 0x7E, 0x00 },
  { 0x1C, 0x00, 0x78, 0xCC, 0xFC, 0xC0, 0x78, 0x00 },
  { 0x7E, 0xC3, 0x3C, 0x06, 0x3E, 0x66, 0x3F, 0x00 },
  { 0xCC, 0x00, 0x78, 0x0C, 0x7C, 0xCC, 0x7E, 0x00 },
  { 0xE0, 0x00, 0x78, 0x0C, 0x7C, 0xCC, 0x7E, 0x00 },
  { 0x30, 0x30, 0x78, 0x0C, 0x7C, 0xCC, 0x7E, 0x00 },
  { 0x00, 0x00, 0x78, 0xC0, 0xC0, 0x78, 0x0C, 0x38 },
  { 0x7E, 0xC3, 0x3C, 0x66, 0x7E, 0x60, 0x3C, 0x00 },
  { 0xCC, 0x00, 0x78, 0xCC, 0xFC, 0xC0, 0x78, 0x00 },
  { 0xE0, 0x00, 0x78, 0xCC, 0xFC, 0xC0, 0x78, 0x00 },
  { 0xCC, 0x00, 0x70, 0x30, 0x30, 0x30, 0x78, 0x00 },
  { 0x7C, 0xC6, 0x38, 0x18, 0x18, 0x18, 0x3C, 0x00 },
  { 0xE0, 0x00, 0x70, 0x30, 0x30, 0x30, 0x78, 0x00 },
  { 0xC6, 0x38, 0x6C, 0xC6, 0xFE, 0xC6, 0xC6, 0x00 },
  { 0x30, 0x30, 0x00, 0x78, 0xCC, 0xFC, 0xCC, 0x00 },
  { 0x1C, 0x00, 0xFC, 0x60, 0x78, 0x60, 0xFC, 0x00 },
  { 0x00, 0x00, 0x7F, 0x0C, 0x7F, 0xCC, 0x7F, 0x00 },
  { 0x3E, 0x6C, 0xCC, 0xFE, 0xCC, 0xCC, 0xCE, 0x00 },
  { 0x78, 0xCC, 0x00, 0x78, 0xCC, 0xCC, 0x78, 0x00 },
  { 0x00, 0xCC, 0x00, 0x78, 0xCC, 0xCC, 0x78, 0x00 },
  { 0x00, 0xE0, 0x00, 0x78, 0xCC, 0xCC, 0x78, 0x00 },
  { 0x78, 0xCC, 0x00, 0xCC, 0xCC, 0xCC, 0x7E, 0x00 },
  { 0x00, 0xE0, 0x00, 0xCC, 0xCC, 0xCC, 0x7E, 0x00 },
  { 0x00, 0xCC, 0x00, 0xCC, 0xCC, 0x7C, 0x0C, 0xF8 },
  { 0xC3, 0x18, 0x3C, 0x66, 0x66, 0x3C, 0x18, 0x00 },
  { 0xCC, 0x00, 0xCC, 0xCC, 0xCC, 0xCC, 0x78, 0x00 },
  { 0x18, 0x18, 0x7E, 0xC0, 0xC0, 0x7E, 0x18, 0x18 },
  { 0x38, 0x6C, 0x64, 0xF0, 0x60, 0xE6, 0xFC, 0x00 },
  { 0xCC, 0xCC, 0x78, 0xFC, 0x30, 0xFC, 0x30, 0x30 },
  { 0xF8, 0xCC, 0xCC, 0xFA, 0xC6, 0xCF, 0xC6, 0xC7 },
  { 0x0E, 0x1B, 0x18, 0x3C, 0x18, 0x18, 0xD8, 0x70 },
  { 0x1C, 0x00, 0x78, 0x0C, 0x7C, 0xCC, 0x7E, 0x00 },
  { 0x38, 0x00, 0x70, 0x30, 0x30, 0x30, 0x78, 0x00 },
  { 0x00, 0x1C, 0x00, 0x78, 0xCC, 0xCC, 0x78, 0x00 },
  { 0x00, 0x1C, 0x00, 0xCC, 0xCC, 0xCC, 0x7E, 0x00 },
  { 0x00, 0xF8, 0x00, 0xF8, 0xCC, 0xCC, 0xCC, 0x00 },
  { 0xFC, 0x00, 0xCC, 0xEC, 0xFC, 0xDC, 0xCC, 0x00 },
  { 0x3C, 0x6C, 0x6C, 0x3E, 0x00, 0x7E, 0x00, 0x00 },
  { 0x38, 0x6C, 0x6C, 0x38, 0x00, 0x7C, 0x00, 0x00 },
  { 0x30, 0x00, 0x30, 0x60, 0xC0, 0xCC, 0x78, 0x00 },
  { 0x00, 0x00, 0x00, 0xFC, 0xC0, 0xC0, 0x00, 0x00 },
  { 0x00, 0x00, 0x00, 0xFC, 0x0C, 0x0C, 0x00, 0x00 },
  { 0xC3, 0xC6, 0xCC, 0xDE, 0x33, 0x66, 0xCC, 0x0F },
  { 0xC3, 0xC6, 0xCC, 0xDB, 0x37, 0x6F, 0xCF, 0x03 },
  { 0x18, 0x18, 0x00, 0x18, 0x18, 0x18, 0x18, 0x00 },
  { 0x00, 0x33, 0x66, 0xCC, 0x66, 0x33, 0x00, 0x00 },
  { 0x00, 0xCC, 0x66, 0x33, 0x66, 0xCC, 0x00, 0x00 },
  { 0x22, 0x88, 0x22, 0x88, 0x22, 0x88, 0x22, 0x88 },
  { 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA },
  { 0xDB, 0x77, 0xDB, 0xEE, 0xDB, 0x77, 0xDB, 0xEE },
  { 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18 },
  { 0x18, 0x18, 0x18, 0x18, 0xF8, 0x18, 0x18, 0x18 },
  { 0x18, 0x18, 0xF8, 0x18, 0xF8, 0x18, 0x18, 0x18 },
  { 0x36, 0x36, 0x36, 0x36, 0xF6, 0x36, 0x36, 0x36 },
  { 0x00, 0x00, 0x00, 0x00, 0xFE, 0x36, 0x36, 0x36 },
  { 0x00, 0x00, 0xF8, 0x18, 0xF8, 0x18, 0x18, 0x18 },
  { 0x36, 0x36, 0xF6, 0x06, 0xF6, 0x36, 0x36, 0x36 },
  { 0x36, 0x36, 0x36, 0x36, 0x36, 0x36, 0x36, 0x36 },
  { 0x00, 0x00, 0xFE, 0x06, 0xF6, 0x36, 0x36, 0x36 },
  { 0x36, 0x36, 0xF6, 0x06, 0xFE, 0x00, 0x00, 0x00 },
  { 0x36, 0x36, 0x36, 0x36, 0xFE, 0x00, 0x00, 0x00 },
  { 0x18, 0x18, 0xF8, 0x18, 0xF8, 0x00, 0x00, 0x00 },
  { 0x00, 0x00, 0x00, 0x00, 0xF8, 0x18, 0x18, 0x18 },
  { 0x18, 0x18, 0x18, 0x18, 0x1F, 0x00, 0x00, 0x00 },
  { 0x18, 0x18, 0x18, 0x18, 0xFF, 0x00, 0x00, 0x00 },
  { 0x00, 0x00, 0x00, 0x00, 0xFF, 0x18, 0x18, 0x18 },
  { 0x18, 0x18, 0x18, 0x18, 0x1F, 0x18, 0x18, 0x18 },
  { 0x00, 0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, 0x00 },
  { 0x18, 0x18, 0x18, 0x18, 0xFF, 0x18, 0x18, 0x18 },
  { 0x18, 0x18, 0x1F, 0x18, 0x1F, 0x18, 0x18, 0x18 },
  { 0x36, 0x36, 0x36, 0x36, 0x37, 0x36, 0x36, 0x36 },
  { 0x36, 0x36, 0x37, 0x30, 0x3F, 0x00, 0x00, 0x00 },
  { 0x00, 0x00, 0x3F, 0x30, 0x37, 0x36, 0x36, 0x36 },
  { 0x36, 0x36, 0xF7, 0x00, 0xFF, 0x00, 0x00, 0x00 },
  { 0x00, 0x00, 0xFF, 0x00, 0xF7, 0x36, 0x36, 0x36 },
  { 0x36, 0x36, 0x37, 0x30, 0x37, 0x36, 0x36, 0x36 },
  { 0x00, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0x00, 0x00 },
  { 0x36, 0x36, 0xF7, 0x00, 0xF7, 0x36, 0x36, 0x36 },
  { 0x18, 0x18, 0xFF, 0x00, 0xFF, 0x00, 0x00, 0x00 },
  { 0x36, 0x36, 0x36, 0x36, 0xFF, 0x00, 0x00, 0x00 },
  { 0x00, 0x00, 0xFF, 0x00, 0xFF, 0x18, 0x18, 0x18 },
  { 0x00, 0x00, 0x00, 0x00, 0xFF, 0x36, 0x36, 0x36 },
  { 0x36, 0x36, 0x36, 0x36, 0x3F, 0x00, 0x00, 0x00 },
  { 0x18, 0x18, 0x1F, 0x18, 0x1F, 0x00, 0x00, 0x00 },
  { 0x00, 0x00, 0x1F, 0x18, 0x1F, 0x18, 0x18, 0x18 },
  { 0x00, 0x00, 0x00, 0x00, 0x3F, 0x36, 0x36, 0x36 },
  { 0x36, 0x36, 0x36, 0x36, 0xFF, 0x36, 0x36, 0x36 },
  { 0x18, 0x18, 0xFF, 0x18, 0xFF, 0x18, 0x18, 0x18 },
  { 0x18, 0x18, 0x18, 0x18, 0xF8, 0x00, 0x00, 0x00 },
  { 0x00, 0x00, 0x00, 0x00, 0x1F, 0x18, 0x18, 0x18 },
  { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF },
  { 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF },
  { 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0 },
  { 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F },
  { 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00 },
  { 0x00, 0x00, 0x76, 0xDC, 0xC8, 0xDC, 0x76, 0x00 },
  { 0x00, 0x78, 0xCC, 0xF8, 0xCC, 0xF8, 0xC0, 0xC0 },
  { 0x00, 0xFC, 0xCC, 0xC0, 0xC0, 0xC0, 0xC0, 0x00 },
  { 0x00, 0xFE, 0x6C, 0x6C, 0x6C, 0x6C, 0x6C, 0x00 },
  { 0xFC, 0xCC, 0x60, 0x30, 0x60, 0xCC, 0xFC, 0x00 },
  { 0x00, 0x00, 0x7E, 0xD8, 0xD8, 0xD8, 0x70, 0x00 },
  { 0x00, 0x66, 0x66, 0x66, 0x66, 0x7C, 0x60, 0xC0 },
  { 0x00, 0x76, 0xDC, 0x18, 0x18, 0x18, 0x18, 0x00 },
  { 0xFC, 0x30, 0x78, 0xCC, 0xCC, 0x78, 0x30, 0xFC },
  { 0x38, 0x6C, 0xC6, 0xFE, 0xC6, 0x6C, 0x38, 0x00 },
  { 0x38, 0x6C, 0xC6, 0xC6, 0x6C, 0x6C, 0xEE, 0x00 },
  { 0x1C, 0x30, 0x18, 0x7C, 0xCC, 0xCC, 0x78, 0x00 },
  { 0x00, 0x00, 0x7E, 0xDB, 0xDB, 0x7E, 0x00, 0x00 },
  { 0x06, 0x0C, 0x7E, 0xDB, 0xDB, 0x7E, 0x60, 0xC0 },
  { 0x38, 0x60, 0xC0, 0xF8, 0xC0, 0x60, 0x38, 0x00 },
  { 0x78, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0x00 },
  { 0x00, 0xFC, 0x00, 0xFC, 0x00, 0xFC, 0x00, 0x00 },
  { 0x30, 0x30, 0xFC, 0x30, 0x30, 0x00, 0xFC, 0x00 },
  { 0x60, 0x30, 0x18, 0x30, 0x60, 0x00, 0xFC, 0x00 },
  { 0x18, 0x30, 0x60, 0x30, 0x18, 0x00, 0xFC, 0x00 },
  { 0x0E, 0x1B, 0x1B, 0x18, 0x18, 0x18, 0x18, 0x18 },
  { 0x18, 0x18, 0x18, 0x18, 0x18, 0xD8, 0xD8, 0x70 },
  { 0x30, 0x30, 0x00, 0xFC, 0x00, 0x30, 0x30, 0x00 },
  { 0x00, 0x76, 0xDC, 0x00, 0x76, 0xDC, 0x00, 0x00 },
  { 0x38, 0x6C, 0x6C, 0x38, 0x00, 0x00, 0x00, 0x00 },
  { 0x00, 0x00, 0x00, 0x18, 0x18, 0x00, 0x00, 0x00 },
  { 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x00, 0x00 },
  { 0x0F, 0x0C, 0x0C, 0x0C, 0xEC, 0x6C, 0x3C, 0x1C },
  { 0x78, 0x6C, 0x6C, 0x6C, 0x6C, 0x00, 0x00, 0x00 },
  { 0x70, 0x18, 0x30, 0x60, 0x78, 0x00, 0x00, 0x00 },
  { 0x00, 0x00, 0x3C, 0x3C, 0x3C, 0x3C, 0x00, 0x00 },
  { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
};

/*
 * gp_djcon.c ends here
 */
