/*
 * gp_djgpp.c --- DJGPP platform specific functions
 *
 * Copyright (C) 1997, 1998  Michael Bukin
 */

/* DJGPP 2.01 specific routines for Ghostscript.  */

#include "stdio_.h"
#include "string_.h"
#include "dos_.h"
#include "gx.h"
#include "gscdefs.h"
#include "gsexit.h"
#include "gstypes.h"
#include "gsmemory.h"
#include "gsstruct.h"
#include "gsutil.h"
#include "gp.h"

#include <ctype.h>
#include <dpmi.h>
#include <fcntl.h>
#include <glob.h>
#include <io.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/* ------ Initialization/termination ------ */

/*
 * This routine is called early in the initialization.
 * It should do as little as possible.  In particular, it should not
 * do things like open display connections: that is the responsibility
 * of the display device driver.
 */
void
gp_init (void)
{
  static const char *gspath = "GS_LIB=%s/share/gs/%d.%02d;%s/share/gs/fonts";

  /* If $GS_LIB is not there, create it from $DJDIR.  */
  if (getenv ("GS_LIB") == 0)
    {
      char* djdir = getenv ("DJDIR");

      if (djdir != 0)
	{
	  char* gslib;

	  gslib = (char*) malloc (strlen (djdir) * 2 + strlen (gspath) + 10);
	  if (gslib != 0)
	    {
	      sprintf (gslib, gspath,
		       djdir, (int) (gs_revision / 100),
		       (int) (gs_revision % 100), djdir);
	      putenv (gslib);
	      free (gslib);
	    }
	}
    }
}

/*
 * This routine is called just before the program exits (normally or
 * abnormally).  It too should do as little as possible.
 */
void
gp_exit (int exit_status, int code)
{
}

/*
 * Exit the program.  Normally this just calls the `exit' library procedure,
 * but it does something different on a few platforms.
 */
void
gp_do_exit (int exit_status)
{
  exit (exit_status);
}

/* ------ Miscelaneous ------ */

/*
 * Get the string corresponding to an OS error number.
 * If no string is available, return NULL.  The caller may assume
 * the string is allocated statically and permanently.
 */
const char*
gp_strerror (int errnum)
{
  return strerror (errnum);
}

/* ------ Date and time ------ */

/*
 * Read the current date (in days since Jan. 1, 1980)
 * and time (in milliseconds since midnight).
 */
void
gp_get_clock (long* pdt)
{
  long secs_since_1980;
  time_t tsec;
  struct timeval tp;
  struct timezone tzp;
  struct tm* tm;
  
  if (gettimeofday (&tp, &tzp) == -1)
    {
      lprintf ("Ghostscript: gettimeofday failed!\n");
      gs_exit (1);
    }
  secs_since_1980 = tp.tv_sec - (long) (60 * 60 * 24 * 365.25 * 10);
  tsec = tp.tv_sec;
  tm = localtime (&tsec);
  if (tm->tm_isdst)
    secs_since_1980 += (60 * 60);

  pdt[0] = secs_since_1980 / (60 * 60 * 24);
  pdt[1] = (secs_since_1980 % (60 * 60 * 24)) * 1000;
}

/*
 * Read the current time (in seconds since an implementation-defined epoch)
 * into ptm[0], and fraction (in nanoseconds) into ptm[1].
 */
/* This is for Aladdin Ghostscript 5.03 (and maybe some other versions)  */
void
gp_get_realtime (long* pdt)
{
  struct timeval tp;
  struct timezone tzp;

  if (gettimeofday (&tp, &tzp) == -1)
    {
      lprintf ("Ghostscript: gettimeofday failed!\n");
      gs_exit (1);
    }
  pdt[0] = tp.tv_sec;
  pdt[1] = tp.tv_usec * 1000;
}

/*
 * Read the current user CPU time (in seconds) into ptm[0],
 * and fraction (in nanoseconds) into ptm[1].
 */
/* This is for Aladdin Ghostscript 5.03 (and maybe some other versions)  */
void
gp_get_usertime (long* pdt)
{
  /* Use an approximation.  */
  gp_get_realtime (pdt);
}

/* ------ Screen management ------ */

/* Initialize the console. */
void
gp_init_console (void)
{
}

/* Write a string to the console. */
void
gp_console_puts (const char* str, uint size)
{
}

/* Make the console current on the screen. */
int
gp_make_console_current (gx_device* dev)
{
  return 0;
}

/* Make the graphics current on the screen. */
int
gp_make_graphics_current (gx_device* dev)
{
  return 0;
}

/* Get the environment variable that specifies the display to use. */
const char*
gp_getenv_display (void)
{
  return NULL;
}

/* ------ Printer accessing ------ */

/*
 * Open a connection to a printer.  A null file name means use the
 * standard printer connected to the machine, if any.
 * If possible, support "|command" for opening an output pipe.
 * Return NULL if the connection could not be opened.
 */
FILE*
gp_open_printer (char* fname, int binary_mode)
{
  if ((fname == 0) || (strlen (fname) == 0))
    return fopen ("prn", (binary_mode ? "wb" : "w"));
  else if (fname[0] == '|')
    return popen (fname + 1, (binary_mode ? "wb" : "w"));
  else
    return fopen (fname, (binary_mode ? "wb" : "w"));
}

/*
 * Close the connection to the printer.
 */
void
gp_close_printer (FILE* pfile, const char* fname)
{
  if (fname[0] == '|')
    pclose (pfile);
  else
    fclose (pfile);
}

/* ------ File naming and accessing ------ */

/* Define the character used for separating file names in a list. */
const char gp_file_name_list_separator = ';';

/* Define the default scratch file name prefix. */
const char gp_scratch_file_name_prefix[] = "gs_";

/* Define the name of the null output file. */
const char gp_null_file_name[] = "nul";

/* Define the name that designates the current directory. */
const char gp_current_directory_name[] = ".";

/*
 * Define the string to be concatenated with the file mode
 * for opening files without end-of-line conversion.
 * This is always either "" or "b".
 */
const char gp_fmode_binary_suffix[] = "b";

/*
 * Define the file modes for binary reading or writing.
 * (This is just a convenience: they are "r" or "w" + the suffix.)
 */
const char gp_fmode_rb[] = "rb";
const char gp_fmode_wb[] = "wb";

/*
 * Create and open a scratch file with a given name prefix.
 * Write the actual file name at fname.
 */
FILE*
gp_open_scratch_file (const char* prefix, char* fname, const char* mode)
{
  FILE* file;
  
  /* Very simple (and should delete temporary files on exit).  */
  file = tmpfile ();
  setmode (fileno (file), ((strchr (mode, 'b') != 0) ? O_BINARY : O_TEXT));
  return file;
}

/*
 * Open a file with the given name, as a stream of uninterpreted bytes.
 */
FILE*
gp_fopen (const char* fname, const char* mode)
{
  return fopen (fname, mode);
}

/*
 * Answer whether a file name contains a directory/device specification,
 * i.e. is absolute (not directory- or device-relative).
 */
bool
gp_file_name_is_absolute (const char* fname, uint len)
{
  /*
   * In UNIX, AFAIK, absolute path starts with '/' or '~'.
   * Here (in Ghostscript), it is something different.
   * I guess, it is absolute if drive and/or directory part is specified.
   * It is not absolute if it includes only file name (no path).
   */
  int i;

  /* Path includes drive specification?  */
  if ((len >= 2) && (fname[1] == ':'))
    return true;
  
  /* Path includes directory delimiter?  */
  for (i = 0; i < len; i++)
    if ((fname[i] == '/') || (fname[i] == '\\'))
      return true;
  
  /* Path is not absolute.  */
  return false;
}

/*
 * Answer the string to be used for combining a directory/device prefix
 * with a base file name.  The file name is known to not be absolute.
 */
const char*
gp_file_name_concat_string (const char* prefix, uint plen,
			    const char* fname, uint len)
{
  if (plen > 0)
    switch (prefix[plen - 1])
      {
      case ':': case '/': case '\\':
	return "";
      }
  return "/";
}

/* ------ File enumeration ------ */

struct file_enum_s
{
  char* pat;		/* Pattern string.  */
  int curr;		/* Current file index.  */
  glob_t glob;		/* Result of globing.  */
  gs_memory_t* mem;	/* Memory handler.  */
};

gs_private_st_ptrs1 (st_file_enum, file_enum, "file_enum",
		     file_enum_enum_ptrs, file_enum_reloc_ptrs, pat);

/*
 * Begin an enumeration.  pat is a C string that may contain *s or ?s.
 * The implementor should copy the string to a safe place.
 * If the operating system doesn't support correct, arbitrarily placed
 * *s and ?s, the implementation should modify the string so that it
 * will return a conservative superset of the request, and then use
 * the string_match procedure to select the desired subset.  E.g., if the
 * OS doesn't implement ? (single-character wild card), any consecutive
 * string of ?s should be interpreted as *.  Note that \ can appear in
 * the pattern also, as a quoting character.
 */
file_enum*
gp_enumerate_files_init (const char* pat, uint patlen, gs_memory_t* mem)
{
  file_enum* fe;

  /* Allocate file enumeration structure.  */
  fe = (file_enum*) gs_alloc_struct (mem, file_enum, &st_file_enum,
				     "gp_enumerate_files");
  if (fe == 0)
    return 0;

  /* Initialize structure members.  */
  fe->mem = mem;
  fe->glob.gl_pathc = 0;
  fe->glob.gl_pathv = 0;
  fe->glob.gl_offs = 0;
  fe->curr = 0;

  /* Allocate memory for pattern string.  */
  fe->pat = (char*) gs_alloc_bytes (mem, patlen + 1,
				    "gp_enumerate_files(pattern)");
  if (fe->pat == 0)
    {
      gp_enumerate_files_close (fe);
      return 0;
    }

  strncpy (fe->pat, pat, patlen);
  fe->pat[patlen] = '\0';

  /* Match files.  */
  glob (fe->pat, GLOB_MARK | GLOB_NOSORT, 0, &(fe->glob));

  return fe;
}

/* Parameters for string_match.  */
private const string_match_params smp_file = { '*', '?', '\\', true };

/*
 * Return the next file name in the enumeration.  The client passes in
 * a scratch string and a max length.  If the name of the next file fits,
 * the procedure returns the length.  If it doesn't fit, the procedure
 * returns max length +1.  If there are no more files, the procedure
 * returns -1.
 */
uint
gp_enumerate_files_next (file_enum* fe, char* ptr, uint maxlen)
{
  /* Match files again using string_match.  */
  while (fe->curr < fe->glob.gl_pathc)
    {
      if (string_match (fe->glob.gl_pathv[fe->curr],
			strlen (fe->glob.gl_pathv[fe->curr]),
			fe->pat, strlen (fe->pat), &smp_file))
	break;
      else
	fe->curr++;
    }

  /* Any files left?  */
  if (fe->curr >= fe->glob.gl_pathc)
    return ~(uint) 0;

  /* Length is O.K.?  */
  if (maxlen <= strlen (fe->glob.gl_pathv[fe->curr]))
    return maxlen + 1;

  /* Copy filename and increment file index.  */
  strcpy (ptr, fe->glob.gl_pathv[fe->curr++]);

  return strlen (ptr);
}

/*
 * Clean up a file enumeration.  This is only called to abandon
 * an enumeration partway through: ...next should do it if there are
 * no more files to enumerate.  This should deallocate the file_enum
 * structure and any subsidiary structures, strings, buffers, etc.
 */
void
gp_enumerate_files_close (file_enum* fe)
{
  /* Free globbed files.  */
  globfree (&(fe->glob));

  /* Free pattern.  */
  gs_free_object (fe->mem, (byte*) fe->pat, "gp_enumerate_files(pattern)");

  /* Free structure.  */
  gs_free_object (fe->mem, (byte*) fe, "gp_enumerate_files");
}

/*
 * gp_djgpp.c ends here
 */
